#!/usr/bin/python

"""
This script is used to convert hdf5 segmentation remap files
to npy format. Ran Lu's and Nick Turner's MST to remap converter 
generates hdf5s using Julia. The format of these remap files is 
simple. It's an array where the index of the array is the key
and the value is the remap value. If the remap array is [5,2,39],
this means remap segment ID 1 to 5, 2 to 2, and 3 to 39.

This script reformats this array to be zero indexed to suit python
and converts the array to npy format which will be used by the 
WatershedRemapTask in the pipeline.

Example Command Sequence:
  gsutil cp path/to/remap.h5 .
  python remap2npy.py remap.h5 # outputs remap.npy
  gsutil cp -Z remap.npy gs://neuroglancer/DATASET/LAYER/
"""

# This is obsolete

from __future__ import print_function

import sys

import h5py
import numpy as np

if len(sys.argv) == 1:
  print("You must specify a remap .h5 file.")
  sys.exit()

in_file = sys.argv[1]
out_file = in_file.replace('.h5', '')

with h5py.File(in_file,'r') as f:
  arr = f['main'][:]

print(arr.shape, arr.dtype)

# These remap files are often created by julia or
# matlab which assume 1-indexing. Add a new index
# at 0 to realign with python's 0-indexing. 
arr = np.concatenate( (np.array([0]), arr) )
arr = arr.astype(np.uint32)

np.save(out_file, arr)







import pytest

from functools import partial
from builtins import range
from collections import defaultdict
import json
import os.path
import shutil

import numpy as np
from cloudvolume import CloudVolume, EmptyVolumeException
import cloudvolume.lib as lib
from cloudfiles import CloudFiles
import fastremap
from taskqueue import MockTaskQueue, TaskQueue
import tinybrain

from igneous import (
    DownsampleTask, MeshTask, 
    MeshManifestPrefixTask, MeshManifestFilesystemTask,
    QuantizeTask, HyperSquareConsensusTask,
    DeleteTask, BlackoutTask, ContrastNormalizationTask
)
import igneous.task_creation as tc
from igneous.task_creation import create_downsampling_tasks, create_quantized_affinity_info
from igneous.downsample_scales import create_downsample_scales
from .layer_harness import delete_layer, create_layer

@pytest.mark.parametrize("compression_method", ( None, 'gzip', 'br',))
def test_downsample_no_offset(compression_method):
    delete_layer()
    cf, data = create_layer(size=(1024,1024,128,1), offset=(0,0,0))
    cv = CloudVolume(cf.cloudpath)
    assert len(cv.scales) == 1
    assert len(cv.available_mips) == 1

    cv.commit_info()

    tq = MockTaskQueue()
    tasks = create_downsampling_tasks(cf.cloudpath, mip=0, num_mips=4, compress=compression_method)
    tq.insert_all(tasks)

    cv.refresh_info()

    assert len(cv.available_mips) == 5
    assert np.array_equal(cv.mip_volume_size(0), [ 1024, 1024, 128 ])
    assert np.array_equal(cv.mip_volume_size(1), [  512,  512, 128 ])
    assert np.array_equal(cv.mip_volume_size(2), [  256,  256, 128 ])
    assert np.array_equal(cv.mip_volume_size(3), [  128,  128, 128 ])
    assert np.array_equal(cv.mip_volume_size(4), [   64,   64, 128 ])
    
    slice64 = np.s_[0:64, 0:64, 0:64]

    cv.mip = 0
    assert np.all(cv[slice64] == data[slice64])

    data_ds1, = tinybrain.downsample_with_averaging(data, factor=[2, 2, 1, 1])
    cv.mip = 1
    assert np.all(cv[slice64] == data_ds1[slice64])

    data_ds2, = tinybrain.downsample_with_averaging(data, factor=[4, 4, 1, 1])
    cv.mip = 2
    assert np.all(cv[slice64] == data_ds2[slice64])

    data_ds3, = tinybrain.downsample_with_averaging(data, factor=[8, 8, 1, 1])
    cv.mip = 3
    assert np.all(cv[slice64] == data_ds3[slice64])

    data_ds4, = tinybrain.downsample_with_averaging(data, factor=[16, 16, 1, 1])
    cv.mip = 4
    assert np.all(cv[slice64] == data_ds4[slice64])

def test_downsample_no_offset_2x2x2():
    delete_layer()
    cf, data = create_layer(size=(512,512,512,1), offset=(0,0,0))
    cv = CloudVolume(cf.cloudpath)
    assert len(cv.scales) == 1
    assert len(cv.available_mips) == 1

    cv.commit_info()

    tq = MockTaskQueue()
    tasks = create_downsampling_tasks(
        cf.cloudpath, mip=0, num_mips=3, 
        compress=None, factor=(2,2,2)
    )
    tq.insert_all(tasks)

    cv.refresh_info()

    assert len(cv.available_mips) == 4
    assert np.array_equal(cv.mip_volume_size(0), [ 512, 512, 512 ])
    assert np.array_equal(cv.mip_volume_size(1), [ 256, 256, 256 ])
    assert np.array_equal(cv.mip_volume_size(2), [ 128, 128, 128 ])
    assert np.array_equal(cv.mip_volume_size(3), [  64,  64,  64 ])
    
    slice64 = np.s_[0:64, 0:64, 0:64]

    cv.mip = 0
    assert np.all(cv[slice64] == data[slice64])

    data_ds1, = tinybrain.downsample_with_averaging(data, factor=[2, 2, 2, 1])
    cv.mip = 1
    assert np.all(cv[slice64] == data_ds1[slice64])

    data_ds2, = tinybrain.downsample_with_averaging(data, factor=[4, 4, 4, 1])
    cv.mip = 2
    assert np.all(cv[slice64] == data_ds2[slice64])

    data_ds3, = tinybrain.downsample_with_averaging(data, factor=[8, 8, 8, 1])
    cv.mip = 3
    assert np.all(cv[slice64] == data_ds3[slice64])

def test_downsample_with_offset():
    delete_layer()
    cf, data = create_layer(size=(512,512,128,1), offset=(3,7,11))
    cv = CloudVolume(cf.cloudpath)
    assert len(cv.scales) == 1
    assert len(cv.available_mips) == 1

    cv.commit_info()

    tq = MockTaskQueue()
    tasks = create_downsampling_tasks(cf.cloudpath, mip=0, num_mips=3)
    tq.insert_all(tasks)

    cv.refresh_info()

    assert len(cv.available_mips) == 4
    assert np.array_equal(cv.mip_volume_size(0), [ 512, 512, 128 ])
    assert np.array_equal(cv.mip_volume_size(1), [ 256, 256, 128 ])
    assert np.array_equal(cv.mip_volume_size(2), [ 128, 128, 128 ])
    assert np.array_equal(cv.mip_volume_size(3), [  64,  64, 128 ])

    assert np.all(cv.mip_voxel_offset(3) == (0,0,11))
    
    cv.mip = 0
    assert np.all(cv[3:67, 7:71, 11:75] == data[0:64, 0:64, 0:64])

    data_ds1, = tinybrain.downsample_with_averaging(data, factor=[2, 2, 1, 1])
    cv.mip = 1
    assert np.all(cv[1:33, 3:35, 11:75] == data_ds1[0:32, 0:32, 0:64])

    data_ds2, = tinybrain.downsample_with_averaging(data, factor=[4, 4, 1, 1])
    cv.mip = 2
    assert np.all(cv[0:16, 1:17, 11:75] == data_ds2[0:16, 0:16, 0:64])

    data_ds3, = tinybrain.downsample_with_averaging(data, factor=[8, 8, 1, 1])
    cv.mip = 3
    assert np.all(cv[0:8, 0:8, 11:75] == data_ds3[0:8,0:8,0:64])

def test_downsample_w_missing():
    delete_layer()
    cf, data = create_layer(size=(512,512,128,1), offset=(3,7,11))
    cv = CloudVolume(cf.cloudpath)
    assert len(cv.scales) == 1
    assert len(cv.available_mips) == 1
    delete_layer()

    cv.commit_info()

    tq = MockTaskQueue()

    try:
        tasks = create_downsampling_tasks(cf.cloudpath, mip=0, num_mips=3, fill_missing=False)
        tq.insert_all(tasks)
    except EmptyVolumeException:
        pass

    tq = MockTaskQueue()

    tasks = create_downsampling_tasks(cf.cloudpath, mip=0, num_mips=3, fill_missing=True)
    tq.insert_all(tasks)

    cv.refresh_info()

    assert len(cv.available_mips) == 4
    assert np.array_equal(cv.mip_volume_size(0), [ 512, 512, 128 ])
    assert np.array_equal(cv.mip_volume_size(1), [ 256, 256, 128 ])
    assert np.array_equal(cv.mip_volume_size(2), [ 128, 128, 128 ])
    assert np.array_equal(cv.mip_volume_size(3), [  64,  64, 128 ])

    assert np.all(cv.mip_voxel_offset(3) == (0,0,11))
    
    cv.mip = 0
    cv.fill_missing = True
    assert np.count_nonzero(cv[3:67, 7:71, 11:75]) == 0

def test_downsample_higher_mip():
    delete_layer()
    cf, data = create_layer(size=(512,512,64,1), offset=(3,7,11))
    cv = CloudVolume(cf.cloudpath)
    cv.info['scales'] = cv.info['scales'][:1]
    
    tq = MockTaskQueue()

    cv.commit_info()
    tasks = create_downsampling_tasks(cf.cloudpath, mip=0, num_mips=2)
    tq.insert_all(tasks)
    cv.refresh_info()
    assert len(cv.available_mips) == 3

    tasks = create_downsampling_tasks(cf.cloudpath, mip=1, num_mips=2)
    tq.insert_all(tasks)
    cv.refresh_info()
    assert len(cv.available_mips) == 4

    cv.mip = 3
    assert cv[:,:,:].shape == (64,64,64,1)

def test_delete():
    delete_layer()
    cf, _ = create_layer(size=(128,64,64,1), offset=(0,0,0), layer_type="segmentation")
    cv = CloudVolume(cf.cloudpath)

    DeleteTask(
        layer_path=cf.cloudpath,
        offset=(0,0,0),
        shape=(128, 64, 64),
    )

    fnames = list(cf.list())
    
    assert '1_1_1/0-64_0-64_0-64' not in fnames
    assert '1_1_1/64-128_0-64_0-64' not in fnames

def test_blackout_tasks():
    delete_layer()
    cf, _ = create_layer(size=(128,64,64,1), offset=(0,0,0), layer_type="image")
    cv = CloudVolume(cf.cloudpath)

    tq = TaskQueue("fq:///tmp/removeme/blackout/")

    tq.insert(
        partial(BlackoutTask, 
            cloudpath=cf.cloudpath,
            mip=0,
            offset=(0,0,0),
            shape=(128, 64, 64),
            value=11,
            non_aligned_writes=False
        )
    )
    tq.lease().execute()

    img = cv[:,:,:]
    assert np.all(img == 11)

    BlackoutTask(
        cloudpath=cf.cloudpath,
        mip=0,
        offset=(0,0,0),
        shape=(37, 64, 64),
        value=23,
        non_aligned_writes=True
    )

    img = cv[:37,:,:]
    assert np.all(img == 23)

    img = cv[:]
    items, counts = np.unique(img, return_counts=True)
    counts = {
        items[0]: counts[0],
        items[1]: counts[1]
    }

    twenty_threes = 37 * 64 * 64
    assert counts[23] == twenty_threes
    assert counts[11] == (128 * 64 * 64) - twenty_threes
    
@pytest.mark.parametrize('compress', ('gzip', 'br'))
def test_mesh(compress):
    delete_layer()
    cf, _ = create_layer(size=(64,64,64,1), offset=(0,0,0), layer_type="segmentation")
    cv = CloudVolume(cf.cloudpath)
    # create a box of ones surrounded by zeroes
    data = np.zeros(shape=(64,64,64,1), dtype=np.uint32)
    data[1:-1,1:-1,1:-1,:] = 1
    cv[0:64,0:64,0:64] = data
    cv.info['mesh'] = 'mesh'
    cv.commit_info()

    t = MeshTask(
        shape=(64,64,64),
        offset=(0,0,0),
        layer_path=cf.cloudpath,
        mip=0,
        remap_table={"1": "10"},
        low_padding=0,
        high_padding=1,
        compress=compress
    )
    t.execute()
    assert cf.get('mesh/10:0:0-64_0-64_0-64') is not None 
    assert list(cf.list('mesh/')) == ['mesh/10:0:0-64_0-64_0-64']


def test_quantize():
    qpath = 'file:///tmp/removeme/quantized/'

    delete_layer()
    delete_layer(qpath)

    cf, _ = create_layer(size=(256,256,128,3), offset=(0,0,0), layer_type="affinities")
    cv = CloudVolume(cf.cloudpath)

    shape = (128, 128, 64)
    slices = np.s_[ :shape[0], :shape[1], :shape[2], :1 ]

    data = cv[slices]
    data *= 255.0
    data = data.astype(np.uint8)

    task = partial(QuantizeTask,
        source_layer_path=cf.cloudpath,
        dest_layer_path=qpath,
        shape=shape,
        offset=(0,0,0),
        mip=0,
    )

    info = create_quantized_affinity_info(
        cf.cloudpath, qpath, shape, 
        mip=0, chunk_size=[64,64,64], encoding='raw'
    )
    qcv = CloudVolume(qpath, info=info)
    qcv.commit_info()

    create_downsample_scales(qpath, mip=0, ds_shape=shape)

    task()

    qcv.mip = 0

    qdata = qcv[slices]

    assert np.all(data.shape == qdata.shape)
    assert np.all(data == qdata)
    assert data.dtype == np.uint8

def test_mesh_manifests_filesystem():
    directory = '/tmp/removeme/mesh_manifests_fs/'
    layer_path = 'file://' + directory
    mesh_dir = 'mesh_mip_3_error_40'

    delete_layer(layer_path)

    to_path = lambda filename: os.path.join(directory, mesh_dir, filename)

    n_segids = 100
    n_lods = 2
    n_fragids = 5

    CloudFiles(layer_path).put_json('info', {"mesh":"mesh_mip_3_error_40"})

    for segid in range(n_segids):
        for lod in range(n_lods):
            for fragid in range(n_fragids):
                filename = '{}:{}:{}'.format(segid, lod, fragid)
                lib.touch(to_path(filename))

    MeshManifestFilesystemTask(layer_path=layer_path, lod=0)

    for segid in range(n_segids):
        for fragid in range(n_fragids):
            filename = '{}:0'.format(segid)
            assert os.path.exists(to_path(filename)), filename
            filename = '{}:1'.format(segid)
            assert not os.path.exists(to_path(filename)), filename

    MeshManifestFilesystemTask(layer_path=layer_path, lod=1)

    for segid in range(n_segids):
        for fragid in range(n_fragids):
            filename = '{}:0'.format(segid)
        assert os.path.exists(to_path(filename)), filename
        filename = '{}:1'.format(segid)
        assert os.path.exists(to_path(filename)), filename

    with open(to_path('50:0'), 'r') as f:
        content = json.loads(f.read())
        content["fragments"].sort()
        assert content == {"fragments": [ "50:0:0","50:0:1","50:0:2","50:0:3","50:0:4" ]}

    if os.path.exists(directory):
        shutil.rmtree(directory)

def test_mesh_manifests_prefixes():
    directory = '/tmp/removeme/mesh_manifests_prefix/'
    layer_path = 'file://' + directory
    mesh_dir = 'mesh_mip_3_error_40'

    delete_layer(layer_path)

    to_path = lambda filename: os.path.join(directory, mesh_dir, filename)

    n_segids = 100
    n_lods = 2
    n_fragids = 5

    CloudFiles(layer_path).put_json('info', {"mesh":"mesh_mip_3_error_40"})

    for segid in range(n_segids):
        for lod in range(n_lods):
            for fragid in range(n_fragids):
                filename = '{}:{}:{}'.format(segid, lod, fragid)
                lib.touch(to_path(filename))

    for i in range(10):
        MeshManifestPrefixTask(layer_path=layer_path, prefix=str(i), lod=0)

    for segid in range(n_segids):
        for fragid in range(n_fragids):
            filename = '{}:0'.format(segid)
            assert os.path.exists(to_path(filename))
            filename = '{}:1'.format(segid)
            assert not os.path.exists(to_path(filename))

    for i in range(10):
        MeshManifestPrefixTask(layer_path=layer_path, prefix=str(i), lod=1)

    for segid in range(n_segids):
        for fragid in range(n_fragids):
            filename = '{}:0'.format(segid)
        assert os.path.exists(to_path(filename))
        filename = '{}:1'.format(segid)
        assert os.path.exists(to_path(filename))

    with open(to_path('50:0'), 'r') as f:
        content = json.loads(f.read())
        assert content == {"fragments": [ "50:0:0","50:0:1","50:0:2","50:0:3","50:0:4" ]}

    if os.path.exists(directory):
        shutil.rmtree(directory)

def test_luminance_levels_task():
    directory = '/tmp/removeme/luminance_levels/'
    layer_path = 'file://' + directory

    delete_layer(layer_path)

    cf, imgd = create_layer(
        size=(256,256,128,1), offset=(0,0,0), 
        layer_type="image", layer_name='luminance_levels'
    )

    tq = MockTaskQueue()
    tasks = tc.create_luminance_levels_tasks( 
        layer_path=layer_path,
        coverage_factor=0.01, 
        shape=None, 
        offset=(0,0,0), 
        mip=0
    )
    tq.insert_all(tasks)

    gt = [ 0 ] * 256
    for x,y,z in lib.xyzrange( (0,0,0), list(imgd.shape[:2]) + [1] ):
        gt[ imgd[x,y,0,0] ] += 1

    with open('/tmp/removeme/luminance_levels/levels/0/0', 'rt') as f:
        levels = f.read()

    levels = json.loads(levels)
    assert levels['coverage_ratio'] == 1.0
    assert levels['levels'] == gt

def test_contrast_normalization_task():
    directory = '/tmp/removeme/contrast_normalization/'
    src_path = 'file://' + directory
    dest_path = src_path[:-1] + '2'

    delete_layer(src_path)
    delete_layer(dest_path)

    cf, imgd = create_layer(
        size=(300,300,129,1), offset=(0,0,0), 
        layer_type="image", layer_name='contrast_normalization'
    )
    tq = MockTaskQueue()
    tasks = tc.create_luminance_levels_tasks( 
        layer_path=src_path,
        coverage_factor=0.01, 
        shape=None, 
        offset=(0,0,0), 
        mip=0
    )
    tq.insert_all(tasks)

    tasks = tc.create_contrast_normalization_tasks( 
        src_path=src_path, 
        dest_path=dest_path, 
        levels_path=None,
        shape=None, 
        mip=0, 
        clip_fraction=0.01, 
        fill_missing=False, 
        translate=(0,0,0),
        minval=None, 
        maxval=None, 
        bounds=None,
        bounds_mip=0,
    )
    tq.insert_all(tasks)


@pytest.mark.parametrize("cross_sectional_area", [False, True])
def test_skeletonization_task(cross_sectional_area):
    directory = '/tmp/removeme/skeleton/'
    layer_path = 'file://' + directory
    delete_layer(layer_path)

    img = np.ones((128,128,128), dtype=np.uint64)
    img[:,:,:] = 2
    cv = CloudVolume.from_numpy(
        img,
        layer_type='segmentation',
        vol_path=layer_path,
    )

    tq = MockTaskQueue()
    tasks = tc.create_skeletonizing_tasks(
        layer_path,
        mip=0,
        teasar_params={
            'scale': 10,
            'const': 10,
        },
        cross_sectional_area=cross_sectional_area,
        cross_sectional_area_smoothing_window=1,
    )
    tq.insert_all(tasks)

    cv = CloudVolume(layer_path)
    skel = cv.skeleton.get(2)

    assert len(skel.vertices) > 0
    if cross_sectional_area:
        assert len(skel.cross_sectional_area) > 0
        assert np.all(skel.cross_sectional_area >= 0)

def test_voxel_counting_task():
    directory = '/tmp/removeme/voxelcounting/'
    layer_path = 'file://' + directory
    delete_layer(layer_path)

    img = np.random.randint(0, 50, size=(1024,1024,256), dtype=np.uint8)
    cv = CloudVolume.from_numpy(
        img,
        layer_type='segmentation',
        vol_path=layer_path,
    )

    tq = MockTaskQueue()
    tasks = tc.create_voxel_counting_tasks(layer_path, mip=0)
    tq.insert_all(tasks)
    tc.accumulate_voxel_counts(layer_path, mip=0)

    from mapbuffer import IntMap
    im = CloudFiles(layer_path).get(f"{cv.key}/stats/voxel_counts.im")
    im = IntMap(im)
    uniq, cts = fastremap.unique(img, return_counts=True)
    
    cts_dict_gt = { label: ct for label, ct in zip(uniq, cts) }
    cts_dict_task = im.todict()

    assert cts_dict_task == cts_dict_gt


def test_num_mips_from_memory_target():
    from igneous.task_creation.image import num_mips_from_memory_target

    memory = 0
    chunk_size = (128,128,64)
    factor = (2,2,1)
    num_channels = 1

    num_mips = num_mips_from_memory_target(memory, 'uint8', chunk_size, num_channels, factor)
    assert num_mips == 1

    memory = 100e6
    num_mips = num_mips_from_memory_target(memory, 'uint8', chunk_size, num_channels, factor)
    assert num_mips == 3

    memory = 100e6
    num_mips = num_mips_from_memory_target(memory, 'uint16', chunk_size, num_channels, factor)
    assert num_mips == 2

    memory = 100e6
    num_mips = num_mips_from_memory_target(memory, 'uint32', chunk_size, num_channels, factor)
    assert num_mips == 2

    memory = 100e6
    num_mips = num_mips_from_memory_target(memory, 'uint64', chunk_size, num_channels, factor)
    assert num_mips == 1

    memory = 3.5e9
    num_mips = num_mips_from_memory_target(memory, 'uint64', chunk_size, num_channels, factor)
    assert num_mips == 4

    memory = 12e9
    num_mips = num_mips_from_memory_target(memory, 'uint64', chunk_size, num_channels, factor)
    assert num_mips == 5

    factor = (2,2,2)

    memory = 800e6
    num_mips = num_mips_from_memory_target(memory, 'uint8', chunk_size, num_channels, factor)
    assert num_mips == 3

    memory = 500e6
    num_mips = num_mips_from_memory_target(memory, 'uint8', chunk_size, num_channels, factor)
    assert num_mips == 2

    memory = 100e6
    num_mips = num_mips_from_memory_target(memory, 'uint8', chunk_size, num_channels, factor)
    assert num_mips == 2

    memory = 50e6
    num_mips = num_mips_from_memory_target(memory, 'uint8', chunk_size, num_channels, factor)
    assert num_mips == 1



from typing import Tuple, Dict, Any

import numpy as np

Triple = Tuple[int,int,int]

def draco_encoding_settings(
  shape:Triple, offset:Triple, resolution:Triple,
  compression_level:int, create_metadata:bool,
  uses_new_draco_bin_size=False
) -> Dict[str,Any]:
  chunk_offset_nm = offset * resolution
  
  min_quantization_range = max(shape * resolution)
  if uses_new_draco_bin_size:
    max_draco_bin_size = np.floor(min(resolution) / 2)
  else:
    max_draco_bin_size = np.floor(min(resolution) / np.sqrt(2))

  (
    draco_quantization_bits,
    draco_quantization_range,
    draco_bin_size,
  ) = calculate_draco_quantization_bits_and_range(
    min_quantization_range, max_draco_bin_size
  )
  draco_quantization_origin = chunk_offset_nm - (chunk_offset_nm % draco_bin_size)
  return {
    "quantization_bits": draco_quantization_bits,
    "compression_level": compression_level,
    "quantization_range": draco_quantization_range,
    "quantization_origin": draco_quantization_origin,
    "create_metadata": create_metadata,
  }

def calculate_draco_quantization_bits_and_range(
  min_quantization_range:int, 
  max_draco_bin_size:int, 
  draco_quantization_bits=None
) -> tuple:
  """
  Computes draco parameters for integer quantizing the meshes.
  """
  if draco_quantization_bits is None:
    draco_quantization_bits = np.ceil(
      np.log2(min_quantization_range / max_draco_bin_size + 1)
    )
  num_draco_bins = 2 ** draco_quantization_bits - 1
  draco_bin_size = np.ceil(min_quantization_range / num_draco_bins)
  draco_quantization_range = draco_bin_size * num_draco_bins
  if draco_quantization_range < min_quantization_range + draco_bin_size:
    if draco_bin_size == max_draco_bin_size:
      return calculate_draco_quantization_bits_and_range(
        min_quantization_range, max_draco_bin_size, draco_quantization_bits + 1
      )
    else:
      draco_bin_size = draco_bin_size + 1
      draco_quantization_range = draco_quantization_range + num_draco_bins
  return draco_quantization_bits, draco_quantization_range, draco_bin_size


from collections import defaultdict
import json
import math
import os
import random
import re
from typing import Optional, Tuple, Union

import numpy as np
from tqdm import tqdm

from cloudfiles import CloudFiles, CloudFile
import cloudfiles.paths

from cloudvolume import CloudVolume
from cloudvolume.lib import Vec, Bbox, jsonify
import mapbuffer
from mapbuffer import MapBuffer, IntMap
from taskqueue import RegisteredTask, queueable

import DracoPy
import fastremap
import zmesh

from .draco import calculate_draco_quantization_bits_and_range, draco_encoding_settings
from . import mesh_graphene_remap

__all__ = [
  "MeshTask", 
  "GrapheneMeshTask", 
  "MeshManifestPrefixTask", 
  "MeshManifestFilesystemTask",
  "TransferMeshFilesTask", 
  "DeleteMeshFilesTask"
]

class MeshTask(RegisteredTask):
  def __init__(self, shape, offset, layer_path, **kwargs):
    """
    Convert all labels in the specified bounding box into meshes
    via marching cubes and quadratic edge collapse (github.com/seung-lab/zmesh).

    Required:
      shape: (sx,sy,sz) size of task
      offset: (x,y,z) offset from (0,0,0)
      layer_path: neuroglancer/cloudvolume dataset path

    Optional:
      lod: (uint) level of detail to record these meshes at
      mip: (uint) level of the resolution pyramid to download segmentation from
      simplification_factor: (uint) try to reduce the number of triangles in the 
        mesh by this factor (but constrained by max_simplification_error)
      max_simplification_error: The maximum physical distance that
        simplification is allowed to move a triangle vertex by. 
      mesh_dir: which subdirectory to write the meshes to (overrides info file location)
      remap_table: agglomerate segmentation before meshing using { orig_id: new_id }
      generate_manifests: (bool) if it is known that the meshes generated by this 
        task will not be cropped by the bounding box, avoid needing to run a seperate
        MeshManifestTask pass by generating manifests on the spot.

      These two options are used to allow sufficient overlap for trivial mesh stitching
      between adjacent tasks.

        low_padding: (uint) expand the bounding box by this many pixels by subtracting
          this padding from the minimum point of the bounding box on all axes.
        high_padding: (uint) expand the bounding box by this many pixels adding
          this padding to the maximum point of the bounding box on all axes.

      parallel_download: (uint: 1) number of processes to use during the segmentation download
      cache_control: (str: None) specify the cache-control header when uploading mesh files
      dust_threshold: (uint: None) don't bother meshing labels strictly smaller than this number of voxels.
      encoding: (str) 'precomputed' (default) or 'draco'
      draco_compression_level: (uint: 1) only applies to draco encoding
      draco_create_metadata: (bool: False) only applies to draco encoding
      progress: (bool: False) show progress bars for meshing 
      object_ids: (list of ints) if specified, only mesh these ids
      fill_missing: (bool: False) replace missing segmentation files with zeros instead of erroring
      spatial_index: (bool: False) generate a JSON spatial index of which meshes are available in
        a given bounding box. 
      sharded: (bool: False) If True, upload all meshes together as a single mapbuffer 
        fragment file. 
      timestamp: (int: None) (graphene only) use the segmentation existing at this
        UNIX timestamp.
    """
    super(MeshTask, self).__init__(shape, offset, layer_path, **kwargs)
    self.shape = Vec(*shape)
    self.offset = Vec(*offset)
    self.layer_path = layer_path
    self.options = {
      'cache_control': kwargs.get('cache_control', None),
      'draco_compression_level': kwargs.get('draco_compression_level', 1),
      'draco_create_metadata': kwargs.get('draco_create_metadata', False),
      'dust_threshold': kwargs.get('dust_threshold', None),
      'dust_global': kwargs.get('dust_global', False),
      'encoding': kwargs.get('encoding', 'precomputed'),
      'fill_missing': kwargs.get('fill_missing', False),
      'generate_manifests': kwargs.get('generate_manifests', False),
      'high_padding': kwargs.get('high_padding', 1),
      'low_padding': kwargs.get('low_padding', 0),
      'lod': kwargs.get('lod', 0),
      'max_simplification_error': kwargs.get('max_simplification_error', 40),
      'simplification_factor': kwargs.get('simplification_factor', 100),
      'mesh_dir': kwargs.get('mesh_dir', None),
      'frag_path': kwargs.get('frag_path', None),
      'mip': kwargs.get('mip', 0),
      'object_ids': kwargs.get('object_ids', None),
      'parallel_download': kwargs.get('parallel_download', 1),
      'progress': kwargs.get('progress', False),
      'remap_table': kwargs.get('remap_table', None),
      'spatial_index': kwargs.get('spatial_index', False),
      'sharded': kwargs.get('sharded', False),
      'timestamp': kwargs.get('timestamp', None),
      'agglomerate': kwargs.get('agglomerate', True),
      'stop_layer': kwargs.get('stop_layer', 2),
      'compress': kwargs.get('compress', 'gzip'),
      'closed_dataset_edges': kwargs.get('closed_dataset_edges', True),
    }
    supported_encodings = ['precomputed', 'draco']
    if not self.options['encoding'] in supported_encodings:
      raise ValueError('Encoding {} is not supported. Options: {}'.format(
        self.options['encoding'], ', '.join(supported_encodings)
      ))
    self._encoding_to_compression_dict = {
      'precomputed': self.options['compress'],
      'draco': False,
    }

  def execute(self):
    self._volume = CloudVolume(
      self.layer_path, self.options['mip'], bounded=False,
      parallel=self.options['parallel_download'], 
      fill_missing=self.options['fill_missing']
    )
    self._bounds = Bbox(self.offset, self.shape + self.offset)
    self._bounds = Bbox.clamp(self._bounds, self._volume.bounds)

    self.progress = bool(self.options['progress'])

    self._mesher = zmesh.Mesher(self._volume.resolution)

    # Marching cubes loves its 1vx overlaps.
    # This avoids lines appearing between
    # adjacent chunks.
    data_bounds = self._bounds.clone()
    data_bounds.minpt -= self.options['low_padding']
    data_bounds.maxpt += self.options['high_padding']

    self._mesh_dir = self.get_mesh_dir()

    if self.options['encoding'] == 'draco':
      self.draco_encoding_settings = draco_encoding_settings(
        shape=(self.shape + self.options['low_padding'] + self.options['high_padding']),
        offset=self.offset,
        resolution=self._volume.resolution,
        compression_level=self.options["draco_compression_level"],
        create_metadata=self.options['draco_create_metadata'],
        uses_new_draco_bin_size=False,
      )

    # chunk_position includes the overlap specified by low_padding/high_padding
    # agglomerate, timestamp, stop_layer only applies to graphene volumes, 
    # no-op for precomputed
    data = self._volume.download(
      data_bounds, 
      agglomerate=self.options['agglomerate'], 
      timestamp=self.options['timestamp'], 
      stop_layer=self.options['stop_layer']
    )

    if not np.any(data):
      if self.options['spatial_index']:
        self._upload_spatial_index(self._bounds, {})
      return

    left_offset = Vec(0,0,0)
    if self.options["closed_dataset_edges"]:
      data, left_offset = self._handle_dataset_boundary(data, data_bounds)

    data = self._remove_dust(
      data, 
      self.options['dust_threshold'], self.options['dust_global']
    )
    data = self._remap(data)

    if self.options['object_ids']:
      data = fastremap.mask_except(data, self.options['object_ids'], in_place=True)

    data, renumbermap = fastremap.renumber(data, in_place=True)
    renumbermap = { v:k for k,v in renumbermap.items() }

    self._mesher.mesh(data[..., 0].T)
    del data

    self.compute_meshes(renumbermap, left_offset)

  def _handle_dataset_boundary(self, data, bbox):
    """
    This logic is used to add a black border along sides
    of the image that touch the dataset boundary which
    results in the closure of the mesh faces on that side.
    """
    if (
      (not np.any(bbox.minpt == self._volume.bounds.minpt))
      and (not np.any(bbox.maxpt == self._volume.bounds.maxpt))
    ):
      return data, Vec(0,0,0)

    shape = Vec(*data.shape, dtype=np.int64)
    offset = Vec(0,0,0,0)
    for i in range(3):
      if bbox.minpt[i] == self._volume.voxel_offset[i]:
        offset[i] += 1
        shape[i] += 1
      if bbox.maxpt[i] == self._volume.bounds.maxpt[i]:
        shape[i] += 1

    slices = (
      slice(offset.x, offset.x + data.shape[0]),
      slice(offset.y, offset.y + data.shape[1]),
      slice(offset.z, offset.z + data.shape[2]),
    )

    mirror_data = np.zeros(shape, dtype=data.dtype, order="F")
    mirror_data[slices] = data
    if offset[0]:
      mirror_data[0,:,:] = 0
    if offset[1]:
      mirror_data[:,0,:] = 0
    if offset[2]:
      mirror_data[:,:,0] = 0

    return mirror_data, offset[:3]

  def get_mesh_dir(self):
    if self.options['mesh_dir'] is not None:
      return self.options['mesh_dir']
    elif 'mesh' in self._volume.info:
      return self._volume.info['mesh']
    else:
      raise ValueError("The mesh destination is not present in the info file.")

  def _remove_dust(self, data, dust_threshold, dust_global):
    if not dust_threshold:
      return data

    if not dust_global:
      segids, pxct = fastremap.unique(data, return_counts=True)
      dust_segids = [ sid for sid, ct in zip(segids, pxct) if ct < int(dust_threshold) ]
      return fastremap.mask(data, dust_segids, in_place=True)
    else:
      return self.apply_dust_global_threshold(dust_threshold, data)

  def apply_dust_global_threshold(self, dust_threshold, all_labels):
    path = self._volume.meta.join(self._volume.cloudpath, self._volume.key, 'stats', 'voxel_counts.im')
    cf = CloudFile(path)
    memcf = CloudFile(path.replace(f"{cf.protocol}://", "mem://"))
    
    if not cf.exists():
      raise FileNotFoundError(f"Cannot apply global dust threshold without {path}")

    buf = None
    if memcf.exists():
      buf = memcf.get()
    else:
      cloudfiles.clear_memory()

    if buf is None:
      if cf.protocol != "file":
        buf = cf.get()
        memcf.put(buf, compress='zstd')
      else:
        buf = cf

    mb = IntMap(buf)
    uniq = fastremap.unique(all_labels)

    valid_objects = []
    for label in uniq:
      if label == 0:
        continue
      if mb[label] >= self.dust_threshold:
        valid_objects.append(label)

    return fastremap.mask_except(all_labels, valid_objects)

  def _remap(self, data):
    if self.options['remap_table'] is None:
      return data 

    self.options['remap_table'] = {
      int(k): int(v) for k, v in self.options['remap_table'].items()
    }

    remap = self.options['remap_table']
    remap[0] = 0

    data = fastremap.mask_except(data, list(remap.keys()), in_place=True)
    return fastremap.remap(data, remap, in_place=True)

  def compute_meshes(self, renumbermap, offset):
    bounding_boxes = {}
    meshes = {}

    for obj_id in tqdm(self._mesher.ids(), disable=(not self.progress), desc="Mesh"):
      remapped_id = renumbermap[obj_id]
      mesh_binary, mesh_bounds = self._create_mesh(obj_id, offset)
      bounding_boxes[remapped_id] = mesh_bounds.to_list()
      meshes[remapped_id] = mesh_binary

    if self.options['sharded']:
      self._upload_batch(meshes, self._bounds)
    else:
      self._upload_individuals(meshes, self.options['generate_manifests'])

    if self.options['spatial_index']:
      self._upload_spatial_index(self._bounds, bounding_boxes)

  def _upload_batch(self, meshes, bbox):
    frag_path = self.options['frag_path'] or self.layer_path
    cf = CloudFiles(frag_path, progress=self.options['progress'])

    mbuf = MapBuffer(meshes, compress="br")

    cf.put(
      f"{self._mesh_dir}/{bbox.to_filename()}.frags",
      content=mbuf.tobytes(),
      compress=None,
      content_type="application/x.mapbuffer",
      cache_control=False,
    )

  def _upload_individuals(self, mesh_binaries, generate_manifests):
    cf = CloudFiles(self.layer_path)

    content_type = "model/mesh"
    if self.options["encoding"] == "draco":
      content_type = "model/x.draco"
    
    cf.puts(
      ( 
        (
          f"{self._mesh_dir}/{segid}:{self.options['lod']}:{self._bounds.to_filename()}", 
          mesh_binary
        ) 
        for segid, mesh_binary in mesh_binaries.items() 
      ),
      compress=self._encoding_to_compression_dict[self.options['encoding']],
      cache_control=self.options['cache_control'],
      content_type=content_type,
    )

    if generate_manifests:
      cf.put_jsons(
        (
          (
            f"{self._mesh_dir}/{segid}:{self.options['lod']}", 
            { "fragments": [ f"{segid}:{self.options['lod']}:{self._bounds.to_filename()}" ] }
          )
          for segid, mesh_binary in mesh_binaries.items()
        ),
        compress=None,
        cache_control=self.options['cache_control'],
      )

  def _create_mesh(self, obj_id, left_bound_offset):
    mesh = self._mesher.get_mesh(
      obj_id,
      simplification_factor=self.options['simplification_factor'],
      max_simplification_error=self.options['max_simplification_error'],
      voxel_centered=True,
    )

    self._mesher.erase(obj_id)

    resolution = self._volume.resolution
    offset = (self._bounds.minpt - self.options['low_padding']).astype(np.float32)
    mesh.vertices[:] += (offset - left_bound_offset) * resolution

    mesh_bounds = Bbox(
      np.amin(mesh.vertices, axis=0), 
      np.amax(mesh.vertices, axis=0)
    )

    if self.options['encoding'] == 'draco':
      mesh_binary = DracoPy.encode(
        mesh.vertices, mesh.faces, 
        **self.draco_encoding_settings
      )
    elif self.options['encoding'] == 'precomputed':
      mesh_binary = mesh.to_precomputed()

    return mesh_binary, mesh_bounds

  def _upload_spatial_index(self, bbox, mesh_bboxes):
    cf = CloudFiles(self.layer_path, progress=self.options['progress'])
    precision = self._volume.mesh.spatial_index.precision
    resolution = self._volume.resolution 

    bbox = bbox.astype(resolution.dtype) * resolution

    cf.put_json(
      f"{self._mesh_dir}/{bbox.to_filename(precision)}.spatial",
      mesh_bboxes,
      compress=self.options['compress'],
      cache_control=False,
    )

class GrapheneMeshTask(RegisteredTask):
  def __init__(self, cloudpath, shape, offset, mip, **kwargs):
    """
    Convert all labels in the specified bounding box into meshes
    via marching cubes and quadratic edge collapse (github.com/seung-lab/zmesh).

    Required:
      shape: (sx,sy,sz) size of task
      offset: (x,y,z) offset from (0,0,0)
      cloudpath: neuroglancer/cloudvolume dataset path

    Optional:
      mip: (uint) level of the resolution pyramid to download segmentation from
      simplification_factor: (uint) try to reduce the number of triangles in the 
        mesh by this factor (but constrained by max_simplification_error)
      max_simplification_error: The maximum physical distance that
        simplification is allowed to move a triangle vertex by. 
      mesh_dir: which subdirectory to write the meshes to (overrides info file location)

      parallel_download: (uint: 1) number of processes to use during the segmentation download
      cache_control: (str: None) specify the cache-control header when uploading mesh files
      dust_threshold: (uint: None) don't bother meshing labels strictly smaller than this number of voxels.
      encoding: (str) 'precomputed' (default) or 'draco'
      draco_compression_level: (uint: 1) only applies to draco encoding
      progress: (bool: False) show progress bars for meshing 
      object_ids: (list of ints) if specified, only mesh these ids
      fill_missing: (bool: False) replace missing segmentation files with zeros instead of erroring
      timestamp: (int: None) (graphene only) use the segmentation existing at this
        UNIX timestamp.
    """
    super(GrapheneMeshTask, self).__init__(cloudpath, shape, offset, mip, **kwargs)
    self.shape = Vec(*shape)
    self.offset = Vec(*offset)
    self.mip = int(mip)
    self.cloudpath = cloudpath
    self.layer_id = 2
    self.overlap_vx = 1
    self.options = {
      'cache_control': kwargs.get('cache_control', None),
      'draco_compression_level': kwargs.get('draco_compression_level', 1),
      'fill_missing': kwargs.get('fill_missing', False),
      'max_simplification_error': kwargs.get('max_simplification_error', 40),
      'simplification_factor': kwargs.get('simplification_factor', 100),
      'mesh_dir': kwargs.get('mesh_dir', None),
      'progress': kwargs.get('progress', False),
      'timestamp': kwargs.get('timestamp', None),
    }

  def execute(self):
    self.cv = CloudVolume(
      self.cloudpath, mip=self.mip, bounded=False,
      fill_missing=self.options['fill_missing'],
      mesh_dir=self.options['mesh_dir'],
    )

    if self.cv.mesh.meta.is_sharded() == False:
      raise ValueError("The mesh sharding parameter must be defined.")

    self.bounds = Bbox(self.offset, self.shape + self.offset)
    self.bounds = Bbox.clamp(self.bounds, self.cv.bounds)

    self.progress = bool(self.options['progress'])

    self.mesher = zmesh.Mesher(self.cv.resolution)

    # Marching cubes needs 1 voxel overlap to properly 
    # stitch adjacent meshes.
    # data_bounds = self.bounds.clone()
    # data_bounds.maxpt += self.overlap_vx

    self.mesh_dir = self.get_mesh_dir()
    self.draco_encoding_settings = draco_encoding_settings(
      shape=(self.shape + self.overlap_vx),
      offset=self.offset,
      resolution=self.cv.resolution,
      compression_level=1,
      create_metadata=True,
      uses_new_draco_bin_size=self.cv.meta.uses_new_draco_bin_size,
    )

    chunk_pos = self.cv.meta.point_to_chunk_position(self.bounds.center(), mip=self.mip)
    
    img = mesh_graphene_remap.remap_segmentation(
      self.cv, 
      chunk_pos.x, chunk_pos.y, chunk_pos.z, 
      mip=self.mip, 
      overlap_vx=self.overlap_vx, 
      time_stamp=self.timestamp,
      progress=self.progress,
    )

    if not np.any(img):
      return

    self.upload_meshes(
      self.compute_meshes(img)
    )

  def get_mesh_dir(self):
    if self.options['mesh_dir'] is not None:
      return self.options['mesh_dir']
    elif 'mesh' in self.cv.info:
      return self.cv.info['mesh']
    else:
      raise ValueError("The mesh destination is not present in the info file.")

  def compute_meshes(self, data):
    data = data.T
    self.mesher.mesh(data)
    del data

    meshes = {}
    for obj_id in tqdm(self.mesher.ids(), disable=(not self.progress), desc="Mesh"):
      # remapped_id = component_map[obj_id]
      meshes[obj_id] = self.create_mesh(obj_id)

    return meshes

  def upload_meshes(self, meshes):
    if len(meshes) == 0:
      return

    reader = self.cv.mesh.readers[self.layer_id] 

    shard_binary = reader.spec.synthesize_shard(meshes)
    # the shard filename is derived from the chunk position,
    # so any label inside this L2 chunk will do
    shard_filename = reader.get_filename(list(meshes.keys())[0]) 

    cf = CloudFiles(self.cv.cloudpath)
    cf.put(
      f"{self.get_mesh_dir()}/initial/{self.layer_id}/{shard_filename}",
      shard_binary,
      compress=None,
      content_type="application/octet-stream",
      cache_control="no-cache",
    )

  def create_mesh(self, obj_id):
    mesh = self.mesher.get_mesh(
      obj_id,
      simplification_factor=self.options['simplification_factor'],
      max_simplification_error=self.options['max_simplification_error'],
      # Graphene meshes were created before we fixed the offset problem
      # so unless otherwise specificed, keep this set to False
      voxel_centered=False, 
    )

    self.mesher.erase(obj_id)
    mesh.vertices[:] += self.bounds.minpt * self.cv.resolution

    mesh_binary = DracoPy.encode(
      mesh.vertices, mesh.faces, 
      **self.draco_encoding_settings
    )

    return mesh_binary

@queueable
def MeshManifestFilesystemTask(
  layer_path:str,
  lod:int = 0,
  mesh_dir:Optional[str] = None,
):
  cf = CloudFiles(layer_path)
  info = cf.get_json('info')

  if mesh_dir is None and 'mesh' in info:
    mesh_dir = info['mesh']

  filepath = cloudfiles.paths.asfilepath(
    cf.join(layer_path, mesh_dir)
  )
  segids = defaultdict(list)

  regexp = re.compile(r'(\d+):(\d+):')
  for entry in os.scandir(filepath):
    if not entry.is_file():
      continue

    filename = os.path.basename(entry.name)
    # `match` implies the beginning (^). `search` matches whole string
    matches = re.search(regexp, filename)

    if not matches:
      continue

    segid, mlod = matches.groups()
    segid, mlod = int(segid), int(mlod)

    if mlod != lod:
      continue

    filename, ext = os.path.splitext(filename)
    segids[segid].append(filename)

  items = ( 
    (
      f"{mesh_dir}/{segid}:{lod}", 
      { "fragments": frags }
    ) 
    for segid, frags in segids.items() 
  )

  cf.put_jsons(items)

@queueable
def MeshManifestPrefixTask(
  layer_path:str, 
  prefix:str,
  lod:int = 0, 
  mesh_dir:Optional[str] = None
):
  """
  Finalize mesh generation by post-processing chunk fragment
  lists into mesh fragment manifests.
  These are necessary for neuroglancer to know which mesh
  fragments to download for a given segid.

  If we parallelize using prefixes single digit prefixes ['0','1',..'9'] all meshes will
  be correctly processed. But if we do ['10','11',..'99'] meshes from [0,9] won't get
  processed and need to be handle specifically by creating tasks that will process
  a single mesh ['0:','1:',..'9:']
  """
  cf = CloudFiles(layer_path)
  info = cf.get_json('info')

  if mesh_dir is None and 'mesh' in info:
    mesh_dir = info['mesh']

  prefix = cf.join(mesh_dir, prefix)
  segids = defaultdict(list)

  regexp = re.compile(r'(\d+):(\d+):')
  for filename in cf.list(prefix=prefix):
    filename = os.path.basename(filename)
    # `match` implies the beginning (^). `search` matches whole string
    matches = re.search(regexp, filename)

    if not matches:
      continue

    segid, mlod = matches.groups()
    segid, mlod = int(segid), int(mlod)

    if mlod != lod:
      continue

    segids[segid].append(filename)

  items = ( 
    (
      f"{mesh_dir}/{segid}:{lod}", 
      { "fragments": frags }
    ) 
    for segid, frags in segids.items() 
  )

  cf.put_jsons(items)

@queueable
def TransferMeshFilesTask(
  src:str,
  dest:str,
  prefix:str,
  mesh_dir:Optional[str] = None
):
  cv_src = CloudVolume(src)
  cv_dest = CloudVolume(dest, mesh_dir=mesh_dir)

  cf_src = CloudFiles(cv_src.mesh.meta.layerpath)
  cf_dest = CloudFiles(cv_dest.mesh.meta.layerpath)

  cf_src.transfer_to(cf_dest, paths=cf_src.list(prefix=prefix))

@queueable
def DeleteMeshFilesTask(
  cloudpath:str,
  prefix:str,
  mesh_dir:Optional[str] = None
):
  cv = CloudVolume(cloudpath, mesh_dir=mesh_dir)
  cf = CloudFiles(cv.mesh.meta.layerpath)
  cf.delete(cf.list(prefix=prefix))




import enum
from typing import Any, Dict, Tuple, Union, Optional

ShapeType = Tuple[int, int, int]

class DownsampleMethods(enum.IntEnum):
  AVERAGE_POOLING = 1
  MODE_POOLING = 2
  MIN_POOLING = 3
  MAX_POOLING = 4
  STRIDING = 5
  AUTO = 6


from typing import Optional, List, Dict, Tuple

from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
import functools
import itertools
import json
import math
import os
import random
import re
import struct

import numpy as np
from tqdm import tqdm

from cloudfiles import CloudFiles, CloudFile

from cloudvolume import CloudVolume, Mesh, paths
from cloudvolume.lib import Vec, Bbox, jsonify, sip, toiter, first
from cloudvolume.datasource.precomputed.mesh.multilod \
  import MultiLevelPrecomputedMeshManifest, to_stored_model_space
from cloudvolume.datasource.precomputed.sharding import synthesize_shard_files

import mapbuffer
from mapbuffer import MapBuffer
from taskqueue import queueable

import DracoPy
import fastremap
import pyfqmr
import trimesh
import zmesh

from .draco import draco_encoding_settings

__all__ = [
  "MultiResShardedMeshMergeTask",
  "MultiResUnshardedMeshMergeTask",
  "MultiResShardedFromUnshardedMeshMergeTask",
]

@queueable
def MultiResUnshardedMeshMergeTask(
  cloudpath:str, 
  prefix:str,
  cache_control:bool = False,
  draco_compression_level:int = 1,
  mesh_dir:Optional[str] = None,
  num_lod:int = 1,
  min_chunk_size:Tuple[int,int,int] = (512,512,512),
  progress:bool = False,
):
  cv = CloudVolume(cloudpath)
  
  if mesh_dir is None and 'mesh' in cv.info:
    mesh_dir = cv.info['mesh']

  files_per_label = get_mesh_filenames_subset(
    cloudpath, mesh_dir, prefix
  )

  cf = CloudFiles(cv.meta.join(cloudpath, mesh_dir))
  for label, filenames in tqdm(files_per_label.items(), disable=(not progress)):
    files = cf.get(filenames)
    # we should handle draco as well
    files = [ Mesh.from_precomputed(f["content"]) for f in files ]
    files = Mesh.concatenate(*files)

    (manifest, mesh) = process_mesh(
      cv, label, files, 
      num_lod, min_chunk_size, 
      draco_compression_level
    )

    if manifest is None:
      continue

    cf.put(f"{label}.index", manifest.to_binary(), cache_control="no-cache")
    cf.put(f"{label}", mesh, cache_control="no-cache")

def process_mesh(
  cv:CloudVolume,
  label:int,
  mesh:Mesh,
  num_lod:int,
  min_chunk_size:Tuple[int,int,int] = (512,512,512),
  draco_compression_level:int = 7,
) -> Tuple[MultiLevelPrecomputedMeshManifest, Mesh]:

  mesh.vertices /= cv.meta.resolution(cv.mesh.meta.mip)

  grid_origin = np.floor(np.min(mesh.vertices, axis=0))
  mesh_shape = (np.max(mesh.vertices, axis=0) - grid_origin).astype(int)

  if np.any(mesh_shape == 0):
    return (None, None)

  min_chunk_size = np.array(min_chunk_size, dtype=int)
  max_lod = int(max(np.min(np.log2(mesh_shape / min_chunk_size)), 0))
  max_lod = min(max_lod, num_lod)

  lods = generate_lods(label, mesh, max_lod)
  chunk_shape = np.ceil(mesh_shape / (2 ** (len(lods) - 1)))

  if np.any(chunk_shape == 0):
    return (None, None)

  lods = [
    create_octree_level_from_mesh(lods[lod], chunk_shape, lod, len(lods)) 
    for lod in range(len(lods)) 
  ]
  fragment_positions = [ nodes for submeshes, nodes in lods ]
  lods = [ submeshes for submeshes, nodes in lods ]

  manifest = MultiLevelPrecomputedMeshManifest(
    segment_id=label,
    chunk_shape=chunk_shape,
    grid_origin=grid_origin, 
    num_lods=len(lods), 
    lod_scales=[ 2**i for i in range(len(lods)) ],
    vertex_offsets=[[0,0,0]] * len(lods),
    num_fragments_per_lod=[ len(lods[lod]) for lod in range(len(lods)) ],
    fragment_positions=fragment_positions,
    fragment_offsets=[], # needs to be set when we have the final value
  )

  vqb = int(cv.mesh.meta.info["vertex_quantization_bits"])

  mesh_binaries = []
  for lod, submeshes in enumerate(lods):
    for frag_no, submesh in enumerate(submeshes):
      submesh.vertices = to_stored_model_space(
        submesh.vertices, manifest, 
        lod=lod,
        vertex_quantization_bits=vqb,
        frag=frag_no,
      )

      minpt = np.min(submesh.vertices, axis=0)
      quantization_range = np.max(submesh.vertices, axis=0) - minpt
      quantization_range = np.max(quantization_range)

      # mesh.vertices must be integer type or mesh will display
      # distorted in neuroglancer.
      try:
        submesh = DracoPy.encode(
          submesh.vertices, submesh.faces, 
          quantization_bits=vqb,
          compression_level=draco_compression_level,
          quantization_range=quantization_range,
          quantization_origin=minpt,
          create_metadata=True,
        )
      except DracoPy.EncodingFailedException:
        submesh = b''

      manifest.fragment_offsets.append(len(submesh))
      mesh_binaries.append(submesh)

  return (manifest, b''.join(mesh_binaries))

def get_mesh_filenames_subset(
  cloudpath:str, mesh_dir:str, prefix:str
):
  prefix = f'{mesh_dir}/{prefix}'
  segids = defaultdict(list)

  cf = CloudFiles(cloudpath)
  meshexpr = re.compile(r'(\d+):(\d+):')
  for filename in cf.list(prefix=prefix):
    filename = os.path.basename(filename)
    # `match` implies the beginning (^). `search` matches whole string
    matches = re.search(meshexpr, filename)

    if not matches:
      continue

    segid, lod = matches.groups()
    segid, lod = int(segid), int(lod)

    if lod != 0:
      continue

    segids[segid].append(filename)

  return segids

@queueable
def MultiResShardedMeshMergeTask(
  cloudpath:str,
  shard_no:str,
  draco_compression_level:int = 1,
  mesh_dir:Optional[str] = None,
  frag_path:Optional[str] = None,
  cache:Optional[bool] = False,
  num_lod:int = 1,
  spatial_index_db:Optional[str] = None,
  min_chunk_size:Tuple[int,int,int] = (128,128,128),
  progress:bool = False
):
  cv = CloudVolume(cloudpath, spatial_index_db=spatial_index_db, cache=cache)
  cv.mip = cv.mesh.meta.mip
  if mesh_dir is None and 'mesh' in cv.info:
    mesh_dir = cv.info['mesh']

  # This looks messy because we are trying to avoid retaining
  # unnecessary memory. In the original skeleton iteration, this was 
  # using 50 GB+ memory on minnie65. So it makes sense to be just
  # as careful with a heavier type of object.
  locations = locations_for_labels(cv, labels_for_shard(cv, shard_no))
  filenames = set(itertools.chain(*locations.values()))
  labels = set(locations.keys())
  del locations
  meshes = collect_mesh_fragments(
    cv, labels, filenames, mesh_dir, frag_path, progress
  )
  del filenames

  # important to iterate this way to avoid
  # creating a copy of meshes vs. { ... for in }
  for label in labels:
    meshes[label] = Mesh.concatenate(*meshes[label])
  del labels

  fname, shard = create_mesh_shard(
    cv, meshes, 
    num_lod, draco_compression_level,
    progress, shard_no, min_chunk_size
  )
  del meshes

  if shard is None:
    return

  cf = CloudFiles(cv.mesh.meta.layerpath)
  cf.put(
    fname, shard,
    compress=False,
    content_type='application/octet-stream',
    cache_control='no-cache',
  )

@queueable
def MultiResShardedFromUnshardedMeshMergeTask(
  src:str,
  dest:str,
  shard_no:str,
  cache_control:bool = False,
  draco_compression_level:int = 1,
  mesh_dir:Optional[str] = None,
  num_lod:int = 1,
  progress:bool = False,
  min_chunk_size:Tuple[int,int,int] = (512,512,512),
):
  cv_src = CloudVolume(src)

  if mesh_dir is None and 'mesh' in cv_src.info:
    mesh_dir = cv_src.info['mesh']

  cv_dest = CloudVolume(dest, mesh_dir=mesh_dir, progress=progress)

  labels = labels_for_shard(cv_dest, shard_no)

  if cv_src.mesh.meta.info['@type'] == "neuroglancer_multilod_draco":
    meshes = cv_src.mesh.get(labels)
  else:
    meshes = cv_src.mesh.get(labels, fuse=False)

  del labels
  
  fname, shard = create_mesh_shard(
    cv_dest, meshes, 
    num_lod, draco_compression_level,
    progress, shard_no, min_chunk_size
  )
  del meshes

  if shard is None:
    return

  cf = CloudFiles(cv_dest.mesh.meta.layerpath)
  cf.put(
    fname, shard, # fname, data
    compress=False,
    content_type='application/octet-stream',
    cache_control='no-cache',
  )

def generate_lods(
  label:int,
  mesh:Mesh,
  num_lods:int,
  decimation_factor:int = 2, 
  aggressiveness:float = 5.5,
  progress:bool = False,
):
  assert num_lods >= 0, num_lods

  lods = [ mesh ]

  # from pyfqmr documentation:
  # threshold = alpha * (iteration + K) ** agressiveness
  # 
  # Threshold is the total error that can be tolerated by
  # deleting a vertex.
  for i in range(1, num_lods+1):
    simplifier = pyfqmr.Simplify()
    simplifier.setMesh(mesh.vertices, mesh.faces)
    simplifier.simplify_mesh(
      target_count=max(int(len(mesh.faces) / (decimation_factor ** i)), 4),
      aggressiveness=aggressiveness,
      preserve_border=True,
      verbose=False,
      # Additional parameters to expose?
      # max_iterations=
      # K=
      # alpha=
      # update_rate=  # Number of iterations between each update.
      # lossless=
      # threshold_lossless=
    )

    lods.append(
      Mesh(*simplifier.getMesh())
    )

  return lods

def create_mesh_shard(
  cv:CloudVolume, meshes:Dict[int, Mesh],
  num_lod:int, draco_compression_level:int,
  progress:bool, shard_no:str, min_chunk_size:Tuple[int,int,int]
):
  meshes = {
    label: process_mesh(
      cv, label, mesh, 
      num_lod, min_chunk_size,
      draco_compression_level=draco_compression_level
    )
    for label, mesh in tqdm(meshes.items(), disable=(not progress))
  }
  data_offset = {
    label: len(manifest)
    for label, (manifest, mesh_binary) in meshes.items()
    if manifest is not None and len(mesh_binary) > 0
  }
  meshes = {
    label: mesh_binary + manifest.to_binary()
    for label, (manifest, mesh_binary) in meshes.items()
    if manifest is not None and len(mesh_binary) > 0
  }

  if len(meshes) == 0:
    return None, None

  shard_files = synthesize_shard_files(
    cv.mesh.reader.spec, meshes, data_offset
  )

  if len(shard_files) != 1:
    raise ValueError(
      "Only one shard file should be generated per task. "
      "Expected: {} Got: {} ".format(
        str(shard_no), ", ".join(shard_files.keys())
    ))

  filename = first(shard_files.keys())
  return filename, shard_files[filename]

def collect_mesh_fragments(
  cv:CloudVolume, 
  labels:List[int], 
  filenames:List[str], 
  mesh_dir:str, 
  frag_path:Optional[str],
  progress:bool = False
) -> Dict[int, List[Mesh]]:
  dirfn = lambda loc: cv.meta.join(mesh_dir, loc)
  filenames = [ dirfn(loc) for loc in filenames ]

  block_size = 20

  if len(filenames) < block_size:
    blocks = [ filenames ]
    n_blocks = 1
  else:
    n_blocks = max(len(filenames) // block_size, 1)
    blocks = sip(filenames, block_size)

  all_meshes = defaultdict(list)
  def process_shardfile(item):
    filename, content = item
    fragment = MapBuffer(content, frombytesfn=Mesh.from_precomputed)

    for label in labels:
      try:
        mesh = fragment[label]
        mesh.id = label
        all_meshes[label].append((filename, mesh))
      except KeyError:
        continue

    if hasattr(content, "close"):
      content.close()

    return filename

  frag_prefix = frag_path or cv.cloudpath
  local_input = False
  if paths.extract(frag_prefix).protocol == "file":
     local_input = True
     frag_prefix = frag_prefix.replace("file://", "", 1)

  for filenames_block in tqdm(blocks, desc="Filename Block", total=n_blocks, disable=(not progress)):
    if local_input:
      all_files = {}
      for filename in filenames_block:
        all_files[filename] = open(os.path.join(frag_prefix, filename), "rb")

      for item in tqdm(all_files.items(), desc="Scanning Fragments", disable=(not progress)):
        process_shardfile(item)
    else:
      all_files = {
        filename: CloudFile(cv.meta.join(frag_prefix, filename), cache_meta=True)
        for filename in filenames_block
      }
      with ThreadPoolExecutor(max_workers=block_size) as executor:
        for filename in executor.map(process_shardfile, all_files.items()):
          pass

  # ensure consistent results across multiple runs
  # by sorting mesh fragments by filename
  for label in all_meshes:
    all_meshes[label].sort(key=lambda pair: pair[0])
    all_meshes[label] = [ pair[1] for pair in all_meshes[label] ]

  return all_meshes

def locations_for_labels(
  cv:CloudVolume, labels:List[int]
) -> Dict[int, List[str]]:

  SPATIAL_EXT = re.compile(r'\.spatial$')
  index_filenames = cv.mesh.spatial_index.file_locations_per_label(labels)
  for label, locations in index_filenames.items():
    for i, location in enumerate(locations):
      bbx = Bbox.from_filename(re.sub(SPATIAL_EXT, '', location))
      bbx /= cv.meta.resolution(cv.mesh.meta.mip)
      index_filenames[label][i] = bbx.to_filename() + '.frags'
  return index_filenames

def labels_for_shard(
  cv:CloudVolume, shard_no:str, progress:bool = False
) -> List[int]:
  """
  Try to fetch precalculated labels from `$shardno.labels` (faster) otherwise, 
  compute which labels are applicable to this shard from the shard index (much slower).
  """
  labels = CloudFiles(cv.mesh.meta.layerpath).get_json(shard_no + '.labels')
  if labels is not None:
    return labels

  labels = cv.mesh.spatial_index.query(cv.bounds * cv.resolution)
  spec = cv.mesh.reader.spec

  return [ 
    lbl for lbl in tqdm(labels, desc="Computing Shard Numbers", disable=(not progress))  \
    if spec.compute_shard_location(lbl).shard_number == shard_no 
  ]
  
## Below functions adapted from 
## https://github.com/google/neuroglancer/issues/272
## Thanks to Hythem Sidky (@hsidky) for sharing his 
## progress and code with the connectomics community.

def cmp_zorder(lhs, rhs) -> bool:
  def less_msb(x: int, y: int) -> bool:
    return x < y and x < (x ^ y)

  # Assume lhs and rhs array-like objects of indices.
  assert len(lhs) == len(rhs)
  # Will contain the most significant dimension.
  msd = 2
  # Loop over the other dimensions.
  for dim in [1, 0]:
    # Check if the current dimension is more significant
    # by comparing the most significant bits.
    if less_msb(lhs[msd] ^ rhs[msd], lhs[dim] ^ rhs[dim]):
      msd = dim
  return lhs[msd] - rhs[msd]

def create_octree_level_from_mesh(mesh, chunk_shape, lod, num_lods):
  """
  Create submeshes by slicing the orignal mesh to produce smaller chunks
  by slicing them from x,y,z dimensions.

  This creates (2^lod)^3 submeshes.
  """
  if lod == num_lods - 1:
    return ([ mesh ], ((0,0,0),) )

  mesh = trimesh.Trimesh(vertices=mesh.vertices, faces=mesh.faces)

  scale = Vec(*(np.array(chunk_shape) * (2 ** lod)))
  offset = Vec(*np.floor(mesh.vertices.min(axis=0)))
  grid_size = Vec(*np.ceil((mesh.vertices.max(axis=0) - offset) / scale), dtype=int)

  nx, ny, nz = np.eye(3)
  ox, oy, oz = offset * np.eye(3)

  submeshes = []
  nodes = []
  for x in range(0, grid_size.x):
    # list(...) required b/c it doesn't like Vec classes
    mesh_x = trimesh.intersections.slice_mesh_plane(mesh, plane_normal=nx, plane_origin=list(nx*x*scale.x+ox))
    mesh_x = trimesh.intersections.slice_mesh_plane(mesh_x, plane_normal=-nx, plane_origin=list(nx*(x+1)*scale.x+ox))
    for y in range(0, grid_size.y):
      mesh_y = trimesh.intersections.slice_mesh_plane(mesh_x, plane_normal=ny, plane_origin=list(ny*y*scale.y+oy))
      mesh_y = trimesh.intersections.slice_mesh_plane(mesh_y, plane_normal=-ny, plane_origin=list(ny*(y+1)*scale.y+oy))
      for z in range(0, grid_size.z):
        mesh_z = trimesh.intersections.slice_mesh_plane(mesh_y, plane_normal=nz, plane_origin=list(nz*z*scale.z+oz))
        mesh_z = trimesh.intersections.slice_mesh_plane(mesh_z, plane_normal=-nz, plane_origin=list(nz*(z+1)*scale.z+oz))

        if len(mesh_z.vertices) == 0:
          continue

        # test for totally degenerate meshes by checking if 
        # all of two axes match, meaning the mesh must be a 
        # point or a line.
        if np.sum([ np.all(mesh_z.vertices[:,i] == mesh_z.vertices[0,i]) for i in range(3) ]) >= 2:
          continue

        submeshes.append(mesh_z)
        nodes.append((x, y, z))

  # Sort in Z-curve order
  submeshes, nodes = zip(
    *sorted(zip(submeshes, nodes),
    key=functools.cmp_to_key(lambda x, y: cmp_zorder(x[1], y[1])))
  )
  # convert back from trimesh to CV Mesh class
  submeshes = [ Mesh(m.vertices, m.faces) for m in submeshes ]

  return (submeshes, nodes)



from functools import reduce
import operator

from cloudvolume import CloudVolume
from cloudvolume.lib import Vec, xyzrange

from igneous.tasks import (
  WatershedRemapTask, 
  MaskAffinitymapTask, InferenceTask
)
from .common import (
  operator_contact, FinelyDividedTaskIterator, 
  get_bounds, num_tasks
)

__all__ = [
  "create_inference_tasks",
  "create_watershed_remap_tasks",
  "create_mask_affinity_map_tasks",
]

def create_inference_tasks(
    image_layer_path, convnet_path, 
    mask_layer_path, output_layer_path, output_block_start, output_block_size, 
    grid_size, patch_size, patch_overlap, cropping_margin_size,
    output_key='output', num_output_channels=3, 
    image_mip=1, output_mip=1, mask_mip=3
  ):
    """
    convnet inference block by block. The block coordinates should be aligned with 
    cloud storage. 
    """
    class InferenceTaskIterator():
      def __len__(self):
        return int(reduce(operator.mul, grid_size))
      def __iter__(self):
        for x, y, z in xyzrange(grid_size):
          output_offset = tuple(s+x*b for (s, x, b) in 
                                zip(output_block_start, (z, y, x), 
                                    output_block_size))
          yield InferenceTask(
              image_layer_path=image_layer_path,
              convnet_path=convnet_path,
              mask_layer_path=mask_layer_path,
              output_layer_path=output_layer_path,
              output_offset=output_offset,
              output_shape=output_block_size,
              patch_size=patch_size, 
              patch_overlap=patch_overlap,
              cropping_margin_size=cropping_margin_size,
              output_key=output_key,
              num_output_channels=num_output_channels,
              image_mip=image_mip,
              output_mip=output_mip,
              mask_mip=mask_mip
          )

        vol = CloudVolume(output_layer_path, mip=output_mip)
        vol.provenance.processing.append({
            'method': {
                'task': 'InferenceTask',
                'image_layer_path': image_layer_path,
                'convnet_path': convnet_path,
                'mask_layer_path': mask_layer_path,
                'output_layer_path': output_layer_path,
                'output_offset': output_offset,
                'output_shape': output_block_size,
                'patch_size': patch_size,
                'patch_overlap': patch_overlap,
                'cropping_margin_size': cropping_margin_size,
                'output_key': output_key,
                'num_output_channels': num_output_channels,
                'image_mip': image_mip,
                'output_mip': output_mip,
                'mask_mip': mask_mip,
            },
            'by': OPERATOR_CONTACT,
            'date': strftime('%Y-%m-%d %H:%M %Z'),
        })
        vol.commit_provenance()

    return InferenceTaskIterator()

def create_watershed_remap_tasks(
    map_path, src_layer_path, dest_layer_path, 
    shape=Vec(2048, 2048, 64)
  ):
  shape = Vec(*shape)
  vol = CloudVolume(src_layer_path)

  downsample_scales.create_downsample_scales(dest_layer_path, mip=0, ds_shape=shape)

  class WatershedRemapTaskIterator(FinelyDividedTaskIterator):
    def task(self, shape, offset):
      return WatershedRemapTask(
        map_path=map_path,
        src_path=src_layer_path,
        dest_path=dest_layer_path,
        shape=shape.clone(),
        offset=offset.clone(),
      )
    
    def on_finish(self):
      dvol = CloudVolume(dest_layer_path)
      dvol.provenance.processing.append({
        'method': {
          'task': 'WatershedRemapTask',
          'src': src_layer_path,
          'dest': dest_layer_path,
          'remap_file': map_path,
          'shape': list(shape),
        },
        'by': OPERATOR_CONTACT,
        'date': strftime('%Y-%m-%d %H:%M %Z'),
      }) 
      dvol.commit_provenance()

  return WatershedRemapTaskIterator(vol.bounds, shape)

# def create_bigarray_task(cloudpath):
#   """
#   Creates one task for each bigarray chunk present in the bigarray folder.
#   These tasks will convert the bigarray chunks into chunks that ingest tasks are able to understand.
#   """
#   class BigArrayTaskIterator():
#     def __iter__(self):    
#       with Storage(cloudpath) as storage:
#         for filename in storage.list_blobs(prefix='bigarray/'):
#           yield BigArrayTask(
#             chunk_path=storage.get_path_to_file('bigarray/'+filename),
#             chunk_encoding='npz', # npz_uint8 to convert affinites float32 affinties to uint8
#             version='{}/{}'.format(storage._path.dataset_name, storage._path.layer_name)
#           )
#   return BigArrayTaskIterator()

def create_mask_affinity_map_tasks(
    aff_input_layer_path, aff_output_layer_path, 
    aff_mip, mask_layer_path, mask_mip, output_block_start, 
    output_block_size, grid_size 
  ):
    """
    affinity map masking block by block. The block coordinates should be aligned with 
    cloud storage. 
    """

    class MaskAffinityMapTaskIterator():
      def __len__(self):
        return int(reduce(operator.mul, grid_size))
      def __iter__(self):
        for x, y, z in xyzrange(grid_size):
          output_bounds = Bbox.from_slices(tuple(slice(s+x*b, s+x*b+b)
                  for (s, x, b) in zip(output_block_start, (z, y, x), output_block_size)))
          yield MaskAffinitymapTask(
              aff_input_layer_path=aff_input_layer_path,
              aff_output_layer_path=aff_output_layer_path,
              aff_mip=aff_mip, 
              mask_layer_path=mask_layer_path,
              mask_mip=mask_mip,
              output_bounds=output_bounds,
          )

        vol = CloudVolume(output_layer_path, mip=aff_mip)
        vol.provenance.processing.append({
            'method': {
                'task': 'InferenceTask',
                'aff_input_layer_path': aff_input_layer_path,
                'aff_output_layer_path': aff_output_layer_path,
                'aff_mip': aff_mip,
                'mask_layer_path': mask_layer_path,
                'mask_mip': mask_mip,
                'output_block_start': output_block_start,
                'output_block_size': output_block_size, 
                'grid_size': grid_size,
            },
            'by': operator_contact(),
            'date': strftime('%Y-%m-%d %H:%M %Z'),
        })
        vol.commit_provenance()

    return MaskAffinityMapTaskIterator()


import copy
from collections import defaultdict
from functools import reduce, partial
import itertools
import re
from typing import (
  Any, Dict, Optional, 
  Union, Tuple, cast,
  Iterator
)

from time import strftime

import numpy as np
from tqdm import tqdm

import cloudvolume
import cloudvolume.exceptions
from cloudvolume import CloudVolume
from cloudvolume.lib import Vec, Bbox, max2, min2, xyzrange, find_closest_divisor, yellow, jsonify
from cloudvolume.datasource.precomputed.sharding import ShardingSpecification
from cloudfiles import CloudFiles
import cloudfiles.paths
import shardcomputer

from igneous.tasks import (
  MeshTask, MeshManifestPrefixTask, 
  MeshManifestFilesystemTask, GrapheneMeshTask,
  SpatialIndexTask, MultiResShardedMeshMergeTask,
  MultiResUnshardedMeshMergeTask, 
  MultiResShardedFromUnshardedMeshMergeTask,
  TransferMeshFilesTask, DeleteMeshFilesTask
)
from .common import (
  operator_contact, FinelyDividedTaskIterator, 
  get_bounds, num_tasks, graphene_prefixes,
  compute_shard_params_for_hashed
)

__all__ = [
  "create_meshing_tasks",
  "create_mesh_manifest_tasks",
  "create_graphene_meshing_tasks",
  "create_graphene_hybrid_mesh_manifest_tasks",
  "create_spatial_index_mesh_tasks",
  "create_unsharded_multires_mesh_tasks",
  "create_sharded_multires_mesh_tasks",
  "create_sharded_multires_mesh_from_unsharded_tasks",
  "create_xfer_meshes_tasks",
  "create_mesh_deletion_tasks",
]

# split the work up into ~1000 tasks (magnitude 3)
def create_mesh_manifest_tasks(layer_path, magnitude=3, mesh_dir=None):
  assert int(magnitude) == magnitude
  assert magnitude >= 0

  protocol = cloudfiles.paths.get_protocol(layer_path)
  if protocol == "file":
    return [
      partial(MeshManifestFilesystemTask, 
        layer_path=layer_path,
        mesh_dir=mesh_dir,
      )
    ]

  start = 10 ** (magnitude - 1)
  end = 10 ** magnitude

  class MeshManifestTaskIterator(object):
    def __len__(self):
      return (10 ** magnitude) - 1
    def __iter__(self):
      for prefix in range(1, start):
        yield partial(MeshManifestPrefixTask, 
          layer_path=layer_path, 
          prefix=str(prefix) + ':', 
          mesh_dir=mesh_dir
        )

      # enumerate from e.g. 100 to 999
      for prefix in range(start, end):
        yield partial(MeshManifestPrefixTask, 
          layer_path=layer_path, 
          prefix=str(prefix),
          mesh_dir=mesh_dir
        )

  return MeshManifestTaskIterator()

def create_mesh_deletion_tasks(
  layer_path:str, 
  magnitude:int = 3, 
  mesh_dir:Optional[str] = None
):
  assert int(magnitude) == magnitude
  assert magnitude >= 0

  cv = CloudVolume(layer_path, mesh_dir=mesh_dir)

  cf = CloudFiles(cv.mesh.meta.layerpath)
  cf.delete('info')

  try:
    if cv.mesh.meta.is_sharded():
      return [ 
        partial(DeleteMeshFilesTask, 
          cloudpath=layer_path,
          prefix="",
          mesh_dir=mesh_dir,
        ) 
      ]

    start = 10 ** (magnitude - 1)
    end = 10 ** magnitude

    class MeshDeleteTaskIterator:
      def __len__(self):
        return (10 ** magnitude)
      def __iter__(self):
        # get spatial index files that start
        # with 0 too.
        yield partial(DeleteMeshFilesTask, 
          cloudpath=layer_path,
          prefix="0",
          mesh_dir=mesh_dir,
        )

        for prefix in range(1, start):
          yield partial(DeleteMeshFilesTask, 
            cloudpath=layer_path, 
            prefix=str(prefix) + ':', 
            mesh_dir=mesh_dir
          )

        # enumerate from e.g. 100 to 999
        for prefix in range(start, end):
          yield partial(DeleteMeshFilesTask, 
            cloudpath=layer_path, 
            prefix=str(prefix),
            mesh_dir=mesh_dir
          )

    return MeshDeleteTaskIterator()
  finally:
    cv.provenance.processing.append({
      'method': {
        'task': 'DeleteMeshFilesTask',
        'layer_path': layer_path,
        'mesh_dir': mesh_dir,
      },
      'by': operator_contact(),
      'date': strftime('%Y-%m-%d %H:%M %Z'),
    }) 
    cv.commit_provenance()


def create_meshing_tasks(
    layer_path, mip, shape=(448, 448, 448), 
    simplification=True, max_simplification_error=40,
    mesh_dir=None, cdn_cache=False, dust_threshold=None,
    object_ids=None, progress=False, fill_missing=False,
    encoding='precomputed', spatial_index=True, frag_path=None, sharded=False,
    compress='gzip', closed_dataset_edges=True, dust_global=False
  ):
  shape = Vec(*shape)

  vol = CloudVolume(layer_path, mip)

  if mesh_dir is None:
    mesh_dir = 'mesh_mip_{}_err_{}'.format(mip, max_simplification_error)

  if not 'mesh' in vol.info:
    vol.info['mesh'] = mesh_dir
    vol.commit_info()

  cf = CloudFiles(layer_path)
  info_filename = '{}/info'.format(mesh_dir)
  mesh_info = cf.get_json(info_filename) or {}
  mesh_info['@type'] = 'neuroglancer_legacy_mesh'
  mesh_info['mip'] = int(vol.mip)
  mesh_info['chunk_size'] = shape.tolist()
  if spatial_index:
    mesh_info['spatial_index'] = {
        'resolution': vol.resolution.tolist(),
        'chunk_size': (shape*vol.resolution).tolist(),
    }
  cf.put_json(info_filename, mesh_info)

  class MeshTaskIterator(FinelyDividedTaskIterator):
    def task(self, shape, offset):
      return MeshTask(
        shape=shape.clone(),
        offset=offset.clone(),
        layer_path=layer_path,
        mip=vol.mip,
        simplification_factor=(0 if not simplification else 100),
        max_simplification_error=max_simplification_error,
        mesh_dir=mesh_dir, 
        cache_control=('' if cdn_cache else 'no-cache'),
        dust_threshold=dust_threshold,
        dust_global=bool(dust_global),
        progress=progress,
        object_ids=object_ids,
        fill_missing=fill_missing,
        encoding=encoding,
        spatial_index=spatial_index,
        frag_path=frag_path,
        sharded=sharded,
        compress=compress,
        closed_dataset_edges=closed_dataset_edges,
      )

    def on_finish(self):
      vol.provenance.processing.append({
        'method': {
          'task': 'MeshTask',
          'layer_path': layer_path,
          'mip': vol.mip,
          'shape': shape.tolist(),
          'simplification': simplification,
          'max_simplification_error': max_simplification_error,
          'mesh_dir': mesh_dir,
          'fill_missing': fill_missing,
          'cdn_cache': cdn_cache,
          'dust_threshold': dust_threshold,
          'encoding': encoding,
          'object_ids': object_ids,
          'spatial_index': spatial_index,
          'frag_path': frag_path,
          'sharded': sharded,
          'compress': compress,
          'closed_dataset_edges': closed_dataset_edges,
          'dust_global': bool(dust_global),
        },
        'by': operator_contact(),
        'date': strftime('%Y-%m-%d %H:%M %Z'),
      }) 
      vol.commit_provenance()

  return MeshTaskIterator(vol.mip_bounds(mip), shape)

def create_graphene_meshing_tasks(
  cloudpath, timestamp, mip,
  simplification=True, max_simplification_error=40,
  mesh_dir=None, cdn_cache=False, object_ids=None, 
  progress=False, fill_missing=False, sharding=None,
  draco_compression_level=1, bounds=None
):
  cv = CloudVolume(cloudpath, mip=mip)

  if mip < cv.meta.watershed_mip:
    raise ValueError("Must mesh at or above the watershed mip level. Watershed MIP: {} Got: {}".format(
      cv.meta.watershed_mip, mip
    ))

  if mesh_dir is None:
    mesh_dir = 'meshes'

  cv.info['mesh'] = mesh_dir # necessary to set the mesh.commit_info() dir right
  if not 'mesh' in cv.info:
    cv.commit_info()

  watershed_downsample_ratio = cv.resolution // cv.meta.resolution(cv.meta.watershed_mip)
  shape = Vec(*cv.meta.graph_chunk_size) // watershed_downsample_ratio

  cv.mesh.meta.info['@type'] = 'neuroglancer_legacy_mesh'
  cv.mesh.meta.info['mip'] = cv.mip
  cv.mesh.meta.info['chunk_size'] = list(shape)
  if sharding:
    cv.mesh.meta.info['sharding'] = sharding
  cv.mesh.meta.commit_info()

  simplification = (0 if not simplification else 100)

  class GrapheneMeshTaskIterator(FinelyDividedTaskIterator):
    def task(self, shape, offset):
      return GrapheneMeshTask(
        cloudpath=cloudpath,
        shape=shape.clone(),
        offset=offset.clone(),
        mip=int(mip),
        simplification_factor=simplification,
        max_simplification_error=max_simplification_error,
        draco_compression_level=draco_compression_level,
        mesh_dir=mesh_dir, 
        cache_control=('' if cdn_cache else 'no-cache'),
        progress=progress,
        fill_missing=fill_missing,
        timestamp=timestamp,
      )

    def on_finish(self):
      cv.provenance.processing.append({
        'method': {
          'task': 'GrapheneMeshTask',
          'cloudpath': cv.cloudpath,
          'shape': cv.meta.graph_chunk_size,
          'mip': int(mip),
          'simplification': simplification,
          'max_simplification_error': max_simplification_error,
          'mesh_dir': mesh_dir,
          'fill_missing': fill_missing,
          'cdn_cache': cdn_cache,
          'timestamp': timestamp,
          'draco_compression_level': draco_compression_level,
        },
        'by': operator_contact(),
        'date': strftime('%Y-%m-%d %H:%M %Z'),
      }) 
      cv.commit_provenance()

  if bounds is None:
    bounds = cv.meta.bounds(mip).clone()
  else:
    bounds = cv.bbox_to_mip(bounds, mip=0, to_mip=mip)
    bounds = Bbox.clamp(bounds, cv.bounds)

  bounds = bounds.expand_to_chunk_size(shape, cv.voxel_offset)

  return GrapheneMeshTaskIterator(bounds, shape)

def create_graphene_hybrid_mesh_manifest_tasks(
  cloudpath, mip, mip_bits, x_bits, y_bits, z_bits
):
  prefixes = graphene_prefixes(mip, mip_bits, (x_bits, y_bits, z_bits))

  class GrapheneHybridMeshManifestTaskIterator(object):
    def __len__(self):
      return len(prefixes)
    def __iter__(self):
      for prefix in prefixes:
        yield partial(MeshManifestPrefixTask, layer_path=cloudpath, prefix=str(prefix))

  return GrapheneHybridMeshManifestTaskIterator()

def create_spatial_index_mesh_tasks(
  cloudpath:str, 
  shape:Tuple[int,int,int] = (448,448,448), 
  mip:int = 0, 
  fill_missing:bool = False, 
  compress:Optional[Union[str,bool]] = 'gzip', 
  mesh_dir:Optional[str] = None
):
  """
  The main way to add a spatial index is to use the MeshTask,
  but old datasets or broken datasets may need it to be 
  reconstituted. An alternative use is create the spatial index
  over a different area size than the mesh task.
  """
  shape = Vec(*shape)

  vol = CloudVolume(cloudpath, mip=mip)

  if mesh_dir is None and not vol.info.get("mesh", None):
    mesh_dir = f"mesh_mip_{mip}_err_40"
  elif mesh_dir is None and vol.info.get("mesh", None):
    mesh_dir = vol.info["mesh"]

  if not "mesh" in vol.info:
    vol.info['mesh'] = mesh_dir
    vol.commit_info()

  cf = CloudFiles(cloudpath)
  info_filename = '{}/info'.format(mesh_dir)
  mesh_info = cf.get_json(info_filename) or {}
  new_mesh_info = copy.deepcopy(mesh_info)
  new_mesh_info['@type'] = new_mesh_info.get('@type', 'neuroglancer_legacy_mesh') 
  new_mesh_info['mip'] = new_mesh_info.get("mip", int(vol.mip))
  new_mesh_info['chunk_size'] = shape.tolist()
  new_mesh_info['spatial_index'] = {
    'resolution': vol.resolution.tolist(),
    'chunk_size': (shape * vol.resolution).tolist(),
  }
  if new_mesh_info != mesh_info:
    cf.put_json(info_filename, new_mesh_info)

  vol = CloudVolume(cloudpath, mip=mip) # reload spatial index

  class SpatialIndexMeshTaskIterator(FinelyDividedTaskIterator):
    def task(self, shape, offset):
      return partial(SpatialIndexTask, 
        cloudpath=cloudpath,
        shape=shape,
        offset=offset,
        subdir=mesh_dir,
        precision=vol.mesh.spatial_index.precision,
        mip=int(mip),
        fill_missing=bool(fill_missing),
        compress=compress,
      )

    def on_finish(self):
      vol.provenance.processing.append({
        'method': {
          'task': 'SpatialIndexTask',
          'cloudpath': vol.cloudpath,
          'shape': shape.tolist(),
          'mip': int(mip),
          'subdir': mesh_dir,
          'fill_missing': fill_missing,
          'compress': compress,
        },
        'by': operator_contact(),
        'date': strftime('%Y-%m-%d %H:%M %Z'),
      }) 
      vol.commit_provenance()

  return SpatialIndexMeshTaskIterator(vol.bounds, shape)

def configure_multires_info(
  cloudpath:str,
  vertex_quantization_bits:int, 
  mesh_dir:str
):
  """
  Computes properties and uploads a multires 
  mesh info file
  """
  assert vertex_quantization_bits in (10, 16), vertex_quantization_bits

  vol = CloudVolume(cloudpath)

  mesh_dir = mesh_dir or vol.info.get("mesh", None)

  if not "mesh" in vol.info:
    vol.info['mesh'] = mesh_dir
    vol.commit_info()

  if vol.mesh.meta.mip is None:
    raise ValueError("Unable to detect mesh resolution. Please specify the resolution in the mesh info file.")

  res = vol.meta.resolution(vol.mesh.meta.mip)
  cf = CloudFiles(cloudpath)
  info_filename = f'{mesh_dir}/info'
  mesh_info = cf.get_json(info_filename) or {}
  new_mesh_info = copy.deepcopy(mesh_info)
  new_mesh_info['@type'] = "neuroglancer_multilod_draco"
  new_mesh_info['vertex_quantization_bits'] = vertex_quantization_bits
  new_mesh_info['transform'] = [ 
    res[0], 0,      0,      0,
    0,      res[1], 0,      0,
    0,      0,      res[2], 0,
  ]
  new_mesh_info['lod_scale_multiplier'] = 1.0

  if new_mesh_info != mesh_info:
    cf.put_json(
      info_filename, new_mesh_info, 
      cache_control="no-cache"
    )

  return new_mesh_info

def create_unsharded_multires_mesh_tasks(
  cloudpath:str, num_lod:int = 0, 
  magnitude:int = 3, mesh_dir:str = None,
  vertex_quantization_bits:int = 16,
  min_chunk_size:Tuple[int,int,int] = (256,256,256),
) -> Iterator:
  """
  vertex_quantization_bits: 10 or 16. Adjusts the precision
    of mesh vertices.
  """
  # split the work up into ~1000 tasks (magnitude 3)
  assert int(magnitude) == magnitude

  configure_multires_info(
    cloudpath, 
    vertex_quantization_bits, 
    mesh_dir
  )

  start = 10 ** (magnitude - 1)
  end = 10 ** magnitude

  class UnshardedMultiResTaskIterator:
    def __len__(self):
      return (10 ** magnitude) - 1
    def __iter__(self):
      for prefix in range(1, start):
        yield partial(MultiResUnshardedMeshMergeTask, 
          cloudpath=cloudpath, 
          prefix=str(prefix) + ':', 
          mesh_dir=mesh_dir,
          num_lod=num_lod,
          min_chunk_size=min_chunk_size,
        )

      # enumerate from e.g. 100 to 999
      for prefix in range(start, end):
        yield partial(MultiResUnshardedMeshMergeTask, 
          cloudpath=cloudpath, 
          prefix=prefix, 
          mesh_dir=mesh_dir,
          num_lod=num_lod,
          min_chunk_size=min_chunk_size,
        )

      self.update_provenance()

    def update_provenance(self):
      cv = CloudVolume(cloudpath)
      cv.provenance.processing.append({
        'method': {
          'task': 'MultiResUnshardedMeshMergeTask',
          'cloudpath': cloudpath,
          'magnitude': int(magnitude),
          'num_lod': int(num_lod),
          'vertex_quantization_bits': int(vertex_quantization_bits),
          'min_chunk_size': tuple(min_chunk_size),
          'mesh_dir': mesh_dir,
        },
        'by': operator_contact(),
        'date': strftime('%Y-%m-%d %H:%M %Z'),
      }) 
      cv.commit_provenance()

  return UnshardedMultiResTaskIterator()

def create_xfer_meshes_tasks(
  src:str,
  dest:str,
  mesh_dir:Optional[str] = None, 
  magnitude=2,
):
  cv_src = CloudVolume(src)
  cf_dest = CloudFiles(dest)

  if not mesh_dir:
    info = cf_dest.get_json("info")
    if info.get("mesh", None):
      mesh_dir = info.get("mesh")

  cf_dest.put_json(f"{mesh_dir}/info", cv_src.mesh.meta.info)

  alphabet = [ str(i) for i in range(10) ]
  if cv_src.mesh.meta.is_sharded():
    alphabet += [ 'a', 'b', 'c', 'd', 'e', 'f' ]

  prefixes = itertools.product(*([ alphabet ] * magnitude))
  prefixes = [ "".join(x) for x in prefixes ]

  # explicitly enumerate all prefixes smaller than the magnitude.
  for i in range(1, magnitude):
    explicit_prefix = itertools.product(*([ alphabet ] * i))
    explicit_prefix = [ "".join(x) for x in explicit_prefix ]
    if cv_src.mesh.meta.is_sharded():
      prefixes += [ f"{x}." for x in explicit_prefix ]
    else:
      prefixes += [ f"{x}:0" for x in explicit_prefix ]

  return [
    partial(TransferMeshFilesTask,
      src=src,
      dest=dest,
      prefix=prefix,
      mesh_dir=mesh_dir,
    )
    for prefix in prefixes
  ]

def create_sharded_multires_mesh_from_unsharded_tasks(
  src:str, 
  dest:str,
  shard_index_bytes=2**13, 
  minishard_index_bytes=2**15,
  min_shards:int = 1,
  num_lod:int = 0, 
  draco_compression_level:int = 1,
  vertex_quantization_bits:int = 16,
  minishard_index_encoding="gzip", 
  mesh_dir:Optional[str] = None, 
  mip:Optional[int] = None,
) -> Iterator[MultiResShardedMeshMergeTask]: 

  cv_src = CloudVolume(src)
  cf = CloudFiles(cv_src.mesh.meta.layerpath)
  
  if not CloudFiles(dest).exists("info"):
    CloudVolume(dest, info=cv_src.info, mesh_dir=mesh_dir).commit_info()

  cv_dest = CloudVolume(dest)
  if cv_dest.mesh.meta.info:
    if cv_dest.mesh.meta.mip is None:
      if mip is not None:
        cv_dest.mesh.meta.mip = mip
      else:
        raise ValueError("mip must be set")
    cv_dest.mesh.meta.commit_info()

  configure_multires_info(
    dest, 
    vertex_quantization_bits, 
    mesh_dir
  )

  all_labels = []
  SEGID_RE = re.compile(r'(\d+)(?:(?::0(?:\.gz|\.br|\.zstd)?$)|\.index$)')
  for path in cf.list():
    match = SEGID_RE.search(path)
    if match is None:
      continue
    (segid,) = match.groups()
    all_labels.append(int(segid))

  (shard_bits, minishard_bits, preshift_bits) = \
    compute_shard_params_for_hashed(
      num_labels=len(all_labels),
      shard_index_bytes=int(shard_index_bytes),
      minishard_index_bytes=int(minishard_index_bytes),
      min_shards=int(min_shards),
    )

  cv_dest = CloudVolume(dest, mesh_dir=mesh_dir)
  if mip is None and cv_src.mesh.meta.mip is not None:
    cv_dest.mesh.meta.info["mip"] = cv_src.mesh.meta.mip
    cv_dest.commit_info()

  spec = ShardingSpecification(
    type='neuroglancer_uint64_sharded_v1',
    preshift_bits=preshift_bits,
    hash='murmurhash3_x86_128',
    minishard_bits=minishard_bits,
    shard_bits=shard_bits,
    minishard_index_encoding=minishard_index_encoding,
    data_encoding="raw", # draco encoded meshes
  )

  cv_dest.mesh.meta.info['sharding'] = spec.to_dict()
  cv_dest.mesh.meta.commit_info()

  cv_dest = CloudVolume(dest, mesh_dir=mesh_dir)

  all_labels = np.fromiter(all_labels, dtype=np.uint64, count=len(all_labels))
  shard_labels = shardcomputer.assign_labels_to_shards(all_labels, preshift_bits, shard_bits, minishard_bits)
  del all_labels

  cf = CloudFiles(cv_dest.mesh.meta.layerpath, progress=True)
  files = ( 
    (str(shardno) + '.labels', labels) 
    for shardno, labels in shard_labels.items() 
  )
  cf.put_jsons(
    files, compress="gzip", 
    cache_control="no-cache", total=len(shard_labels)
  )

  cv_dest.provenance.processing.append({
    'method': {
      'task': 'MultiResShardedFromUnshardedMeshMergeTask',
      'src': src,
      'dest': dest,
      'num_lod': num_lod,
      'vertex_quantization_bits': vertex_quantization_bits,
      'preshift_bits': preshift_bits, 
      'minishard_bits': minishard_bits, 
      'shard_bits': shard_bits,
      'mesh_dir': mesh_dir,
      'draco_compression_level': draco_compression_level,
    },
    'by': operator_contact(),
    'date': strftime('%Y-%m-%d %H:%M %Z'),
  }) 
  cv_dest.commit_provenance()

  return [
    partial(MultiResShardedFromUnshardedMeshMergeTask,
      src=src, 
      dest=dest, 
      shard_no=shard_no, 
      num_lod=num_lod,
      mesh_dir=mesh_dir, 
      draco_compression_level=draco_compression_level,
    )
    for shard_no in shard_labels.keys()
  ]

def create_sharded_multires_mesh_tasks(
  cloudpath:str, 
  shard_index_bytes=2**13, 
  minishard_index_bytes=2**15,
  min_shards:int = 1,
  num_lod:int = 0, 
  draco_compression_level:int = 7,
  vertex_quantization_bits:int = 16,
  minishard_index_encoding="gzip", 
  mesh_dir:Optional[str] = None, 
  spatial_index_db:Optional[str] = None,
  frag_path:Optional[str] = None,
  cache:Optional[bool] = False,
  min_chunk_size:Tuple[int,int,int] = (256,256,256),
  max_labels_per_shard:Optional[int] = None,
) -> Iterator[MultiResShardedMeshMergeTask]: 

  mesh_info = configure_multires_info(
    cloudpath, 
    vertex_quantization_bits, 
    mesh_dir
  )

  # rebuild b/c sharding changes the mesh source class
  cv = CloudVolume(cloudpath, progress=True, spatial_index_db=spatial_index_db) 
  cv.mip = cv.mesh.meta.mip

  # 17 sec to download for pinky100
  all_labels = cv.mesh.spatial_index.query(cv.bounds * cv.resolution)
  
  if max_labels_per_shard is not None:
    assert max_labels_per_shard >= 1
    min_shards = max(int(np.ceil(len(all_labels) / max_labels_per_shard)), min_shards)

  (shard_bits, minishard_bits, preshift_bits) = \
    compute_shard_params_for_hashed(
      num_labels=len(all_labels),
      shard_index_bytes=int(shard_index_bytes),
      minishard_index_bytes=int(minishard_index_bytes),
      min_shards=min_shards,
    )

  spec = ShardingSpecification(
    type='neuroglancer_uint64_sharded_v1',
    preshift_bits=preshift_bits,
    hash='murmurhash3_x86_128',
    minishard_bits=minishard_bits,
    shard_bits=shard_bits,
    minishard_index_encoding=minishard_index_encoding,
    data_encoding="raw", # draco encoded meshes
  )

  cv.mesh.meta.info = mesh_info # ensure no race conditions
  cv.mesh.meta.info['sharding'] = spec.to_dict()
  cv.mesh.meta.commit_info()

  cv = CloudVolume(cloudpath)

  all_labels = np.fromiter(all_labels, dtype=np.uint64, count=len(all_labels))
  shard_labels = shardcomputer.assign_labels_to_shards(all_labels, preshift_bits, shard_bits, minishard_bits)
  del all_labels

  cf = CloudFiles(cv.mesh.meta.layerpath, progress=True)
  files = ( 
    (str(shardno) + '.labels', labels) 
    for shardno, labels in shard_labels.items() 
  )
  cf.put_jsons(
    files, compress="gzip", 
    cache_control="no-cache", total=len(shard_labels)
  )

  cv.provenance.processing.append({
    'method': {
      'task': 'MultiResShardedMeshMergeTask',
      'cloudpath': cloudpath,
      'mip': cv.mesh.meta.mip,
      'num_lod': num_lod,
      'vertex_quantization_bits': vertex_quantization_bits,
      'preshift_bits': preshift_bits, 
      'minishard_bits': minishard_bits, 
      'shard_bits': shard_bits,
      'mesh_dir': mesh_dir,
      'frag_path': frag_path,
      'draco_compression_level': draco_compression_level,
      'min_chunk_size': min_chunk_size,
    },
    'by': operator_contact(),
    'date': strftime('%Y-%m-%d %H:%M %Z'),
  }) 
  cv.commit_provenance()

  return [
    partial(MultiResShardedMeshMergeTask,
      cloudpath, shard_no, 
      num_lod=num_lod,
      mesh_dir=mesh_dir, 
      frag_path=frag_path,
      cache=cache,
      spatial_index_db=spatial_index_db,
      draco_compression_level=draco_compression_level,
      min_chunk_size=min_chunk_size,
    )
    for shard_no in shard_labels.keys()
  ]



#!/usr/bin/env python
import setuptools

setuptools.setup(
  setup_requires=['pbr', 'numpy'],
  entry_points={
    "console_scripts": [
      "igneous=igneous_cli:main"
    ],
  },
  extras_require={
    "mysql": [
      "mysql-connector-python",
    ],
  },
  long_description_content_type="text/markdown",
  pbr=True,
)




import numpy as np

from igneous import downsample_scales


def test_plane_scales_xy():
  scales = downsample_scales.compute_plane_downsampling_scales( 
    (2048, 2048, 512), max_downsampled_size=128
  )

  assert len(scales) == 5
  assert scales[0] == (1,1,1)
  assert scales[1] == (2,2,1)
  assert scales[2] == (4,4,1)
  assert scales[3] == (8,8,1)
  assert scales[4] == (16,16,1)

  scales = downsample_scales.compute_plane_downsampling_scales( 
    (357, 2048, 512), max_downsampled_size=128
  )

  assert len(scales) == 2
  assert scales[0] == (1,1,1)
  assert scales[1] == (2,2,1)

  scales = downsample_scales.compute_plane_downsampling_scales( 
    (0, 2048, 512), max_downsampled_size=128
  )

  assert len(scales) == 1
  assert scales[0] == (1,1,1)


def test_plane_scales_yz():
  scales = downsample_scales.compute_plane_downsampling_scales( 
    (512, 2048, 2048), max_downsampled_size=128, preserve_axis='x'
  )

  assert len(scales) == 5
  assert scales[0] == (1,1,1)
  assert scales[1] == (1,2,2)
  assert scales[2] == (1,4,4)
  assert scales[3] == (1,8,8)
  assert scales[4] == (1,16,16)


  scales = downsample_scales.compute_plane_downsampling_scales( 
    (64, 2048, 2048), max_downsampled_size=128, preserve_axis='x'
  )

  assert len(scales) == 5
  assert scales[0] == (1,1,1)
  assert scales[1] == (1,2,2)
  assert scales[2] == (1,4,4)
  assert scales[3] == (1,8,8)
  assert scales[4] == (1,16,16)


def test_downsample_shape_from_memory_target():
  try:
    downsample_scales.downsample_shape_from_memory_target(1, 0, 1, 1, (2,2,1), 16)
    assert False
  except ValueError:
    pass

  shape = downsample_scales.downsample_shape_from_memory_target(1, 1, 1, 1, (1,1,1), 16)
  assert np.all(shape == (4,4,1))

  shape = downsample_scales.downsample_shape_from_memory_target(2, 1, 1, 1, (1,1,1), 16)
  assert np.all(shape == (2,2,1))

  shape = downsample_scales.downsample_shape_from_memory_target(1, 2, 1, 1, (1,1,1), 16)
  assert np.all(shape == (4,4,1))

  shape = downsample_scales.downsample_shape_from_memory_target(1, 256, 256, 64, (1,1,1), 20e6)
  assert np.all(shape == (512,512,64))

  shape = downsample_scales.downsample_shape_from_memory_target(8, 256, 256, 64, (1,1,1), 35e6)
  assert np.all(shape == (256,256,64))

  shape = downsample_scales.downsample_shape_from_memory_target(8, 256, 256, 64, (1,1,1), 3.5e9)
  assert np.all(shape == (2560,2560,64))

  shape = downsample_scales.downsample_shape_from_memory_target(1, 1, 1, 1, (2,2,1), 16)
  assert np.all(shape == (4,4,1))

  shape = downsample_scales.downsample_shape_from_memory_target(1, 1, 1, 1, (2,2,2), 64)
  assert np.all(shape == (4,4,4))

  try:
    downsample_scales.downsample_shape_from_memory_target(1, 1, 1, 1, (2,2,1), 0)
    assert False
  except ValueError:
    pass

  try:
    downsample_scales.downsample_shape_from_memory_target(8, 128, 128, 64, (2,2,1), 100)
    assert False
  except ValueError:
    pass

  shape = downsample_scales.downsample_shape_from_memory_target(8, 128, 128, 64, (2,2,1), 3.5e9)
  assert np.all(shape == (2048,2048,64))

  shape = downsample_scales.downsample_shape_from_memory_target(1, 128, 128, 64, (2,2,1), 3.5e9)
  assert np.all(shape == (4096,4096,64))

  shape = downsample_scales.downsample_shape_from_memory_target(8, 128, 128, 64, (2,2,2), 3.5e9)
  assert np.all(shape == (512,512,256))

  shape = downsample_scales.downsample_shape_from_memory_target(4, 100, 100, 50, (2,2,1), 3.5e9)
  assert np.all(shape == (3200,3200,50))








from .skeleton import (
  SkeletonTask, UnshardedSkeletonMergeTask, 
  ShardedSkeletonMergeTask, DeleteSkeletonFilesTask,
  TransferSkeletonFilesTask, ShardedFromUnshardedSkeletonMergeTask
)
from .mesh import (
  MeshTask, MeshManifestPrefixTask, 
  MeshManifestFilesystemTask,
  GrapheneMeshTask,
  MultiResUnshardedMeshMergeTask,
  MultiResShardedMeshMergeTask,
  MultiResShardedFromUnshardedMeshMergeTask,
  TransferMeshFilesTask, DeleteMeshFilesTask
)
from .image import (
  HyperSquareConsensusTask, #HyperSquareTask,
  DownsampleTask, QuantizeTask, 
  TransferTask, WatershedRemapTask, DeleteTask, 
  LuminanceLevelsTask, ContrastNormalizationTask,
  MaskAffinitymapTask, InferenceTask, BlackoutTask,
  TouchTask, ImageShardTransferTask, ImageShardDownsampleTask,
  CCLFacesTask, CCLEquivalancesTask, RelabelCCLTask,
  CountVoxelsTask
)
from .spatial_index import SpatialIndexTask